# How to Build Applications Using the Basler blaze SDK

This document describes the settings required for building applications
using the APIs of the Basler blaze SDK for operating Basler blaze-101
cameras.

The `Samples` subdirectory contains a collection of C++, C#, Visual Basic,
Python, and MVTec Halcon sample programs.

Please refer to the [Basler Product Documentation](https://docs.baslerweb.com/blaze-101)
for more information about how to operate a Basler blaze-101 camera and its
features.

## C++

This section provides information about how to set up your build system
for using the Basler blaze SDK in C++ applications.

### CMake

When you use CMake as build system, integrating the Basler blaze SDK is quite
easy.

The Basler blaze SDK provides the `FindBaslerBlaze.cmake` module used to locate
the installation of the Basler blaze SDK. The find module exports the
`BaslerBlaze::BaslerBlazeSDK` CMake target. Pass this target as argument to
`target_link_libraries` and CMake will take care of setting the include path,
 library path, etc.

The location of the find module has to be added to CMake's `CMAKE_MODULE_PATH`.

The following lines serve as an example how this can be done within your
CMakeLists.txt file:

```cmake
if (WIN32 AND CMAKE_SIZEOF_VOID_P EQUAL 8)
  set (BASLER_BLAZE_SDKROOT BASLER_BLAZE_SDKROOT64)
else()
  set (BASLER_BLAZE_SDKROOT BASLER_BLAZE_SDKROOT)
endif()

if (DEFINED ENV{${BASLER_BLAZE_SDKROOT}}) # SDK location given by environment variable
    set(module_path "$ENV{${BASLER_BLAZE_SDKROOT}}/lib/cmake")
elseif(BaslerBlaze_ROOT_DIR) # SDK location specified by CMake variable
    set(module_path ${BaslerBlaze_ROOT_DIR}/lib/cmake)
elseif (NOT WIN32)
     set(module_path "/opt/Basler_blaze/lib/cmake") # recommended install location
endif()

if (EXISTS ${module_path}/FindBaslerBlaze.cmake)
    list(APPEND CMAKE_MODULE_PATH ${module_path})
    unset(module_path)
else()
    message(FATAL_ERROR
        FindBaslerBlaze.cmake module not found. Either set the ${BASLER_BLAZE_SDKROOT} "
            "environment variable or the BaslerBlaze_ROOT_DIR CMake variable to the location "
            "the Basler blaze SDK is installed to.")
endif()
```

Once CMake knows where to find the `FindBaslerBlaze.cmake` file, adding the `BaslerBlazeSDK`
target to your project can be done as illustrated in the following sample:

```cmake
find_package(BaslerBlaze REQUIRED)

# The find module exports the BaslerBlaze::BaslerBlazeSDK target. Pass this target as argument to
# target_link_libraries and CMake will take care of setting the include path, library path, etc.
add_excecutable(FirstSample main.cpp)
target_link_libraries(FirstSample BaslerBlaze::BaslerBlazeSDK)
```

`main.cpp` could look like this:

```c++
#include <iostream>
#include <ConsumerImplHelper/BlazeCamera.h>
using namespace GenTLConsumerImplHelper;
using namespace GenApi;

class Sample {
public:
    int run() {
        m_nBuffersGrabbed = 0;
        try {
            m_Camera.OpenFirstCamera();
            // Enable point cloud data.
            CEnumerationPtr(m_Camera.GetParameter("ComponentSelector"))->FromString("Range");
            CBooleanPtr(m_Camera.GetParameter("ComponentEnable"))->SetValue(true);
            CEnumerationPtr(m_Camera.GetParameter("PixelFormat"))->FromString("Coord3D_ABC32f");
            // Acquire images until the onImageGrabbed callback indicates to stop acquisition.
            m_Camera.GrabContinuous(5, 1000, this, &Sample::onImageGrabbed);
            m_Camera.Close();
        } catch (const GenICam::GenericException& e) {
            std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
            return EXIT_FAILURE;
        }
        return EXIT_SUCCESS;
    }

    bool onImageGrabbed(GrabResult grabResult, BufferParts parts) {
        if (grabResult.status == GrabResult::Timeout) {
            std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
            return false; // Indicate to stop acquisition.
        }
        m_nBuffersGrabbed++;
        if (grabResult.status != GrabResult::Ok) {
            std::cerr << "Image " << m_nBuffersGrabbed << " was not grabbed." << std::endl;
        }
        else {
            // Retrieve the values for the center pixel.
            const int width = (int)parts[0].width; const int height = (int)parts[0].height;
            const int u = (int)(0.5 * width); const int v = (int)(0.5 * height);
            CBlazeCamera::Coord3D* p3DCoordinate = (CBlazeCamera::Coord3D*) parts[0].pData + v * width + u;
            if (p3DCoordinate->z != 0)
                std::cout << "x=" << p3DCoordinate->x << " y=" << p3DCoordinate->y << " z=" << p3DCoordinate->z << "\n";
            else
                std::cout << "x=   n/a y=   n/a z=   n/a\n";
        }
        return m_nBuffersGrabbed < 10; // Indicate to stop acquisition when 10 buffers are grabbed.
    }
private:
    CBlazeCamera m_Camera;
    int          m_nBuffersGrabbed;
};

int main(int argc, char* argv[]) {
    int exitCode = EXIT_SUCCESS;
    try {
        CBlazeCamera::InitProducer();
        exitCode = Sample().run();
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions
    return exitCode;
}
```

### Visual Studio Solutions

When you use Microsoft Visual Studio for building your application, you have
to adjust the project's compiler and linker settings as follows:

Add the following include directory for all project configurations
(Configuration Properties -> C/C++ -> General -> Additional Include Directories):

Platform == Win32:
    `$(BASLER_BLAZE_SDKROOT)\include`

Platform == x64:
    `$(BASLER_BLAZE_SDKROOT64)\include`

Enable runtime type info for Release and Debug project configurations
(Configuration Properties -> C/C++ -> Language -> Enable Runtime Type Info).

Enable C++ exceptions for Release and Debug project configurations
(Configuration Properties -> C/C++ -> Code Generation -> Enable C++ Exceptions).

Add the following library directories for all project configurations
(Configuration Properties -> Linker -> General -> Additional Library Directories):

Platform == Win32:
    `$(BASLER_BLAZE_SDKROOT)\lib\win32`

Platform == x64:
    `$(BASLER_BLAZE_SDKROOT64)\lib\x64`

For all platforms and all project configurations add the following libraries
to the list of delay loaded DLLs
(Configuration Properties -> Linker -> Input -> Delay Loaded DLLs):

`GenApi_MD_VC141_v3_1_Basler_pylon_v5_1.dll;GCBase_MD_VC141_v3_1_Basler_pylon_v5_1.dll`

### GNU make

The following minimalistic Makefile illustrates the compiler and linker
settings required for building a C++ application using the Basler blaze SDK.

```make
BASLER_BLAZE_SDKROOT ?= /opt/Basler_blaze

CXX       ?= g++
RM        ?= rm -rf
CPPFLAGS  := -I$(BASLER_BLAZE_SDKROOT)/include
CXXFLAGS  := -std=c++11 -Wno-unknown-pragmas
LIBDIR    := $(BASLER_BLAZE_SDKROOT)/library/CPP/lib/Linux64_x64
LIBS      := -L$(LIBDIR) -lGenApi_gcc73_v3_1_Basler -lGCBase_gcc73_v3_1_Basler -ldl
GENTL_DIR := $(BASLER_BLAZE_SDKROOT)/lib/gentlproducer/gtl
LDFLAGS   := -Wl,--enable-new-dtags -Wl,-rpath,$(LIBDIR) -Wl,-rpath,$(GENTL_DIR)
# !! The following option is always required when using GenApi. Without
# the -Wl,-E option, runtime type information is not handled properly. Correct
# runtime type information is particularly important for handling exceptions
# thrown by the GenApi library!
LDFLAGS   += -Wl,-E
SRCS      := main.cpp
OBJS      := $(SRCS:.cpp=.o)

FirstSample: $(OBJS)
	$(CXX) $(LDFLAGS) -o FirstSample $(OBJS) $(LIBS)

clean:
	$(RM) $(OBJS) FirstSample
```

## C#/Visual Basic

To operate a Basler blaze-101 camera using C# or Visual Basic .NET, you have to
add the `Basler.blaze.NET` assembly to your project's list of assembly
references.

To to so, select your project in the Visual Studio Solution Explorer and open
the Reference Manager (Main menu -> Project -> Add Reference).

In the dialog's search input field enter `Basler.blazeNET` and select any of
the items listed.

To avoid warnings about potential architecture mismatches when building your
application, it is recommended that you edit the Visual Studio project
file manually to suppress such warnings. In the Solution Explorer, right-click
your project. From the context menu choose the `Unload project` entry.

Open the project's context menu again and select the `Edit project` entry.
The project's XML file is opened in an editor window.

In the XML file navigate to the location where the `Basler.blaze.NET`assembly is
referenced, i.e., search for a fragment similar to the following:

```xml
<ItemGroup>
    <Reference Include="Basler.blaze.NET, Version=1.0.0.0, Culture=neutral,
               PublicKeyToken=def144ed1a820fcc, processorArchitecture=x86" />?
```

Above that `ItemGroup` element add the following XML fragment:

```xml
  <PropertyGroup>
    <ResolveAssemblyWarnOrErrorOnTargetArchitectureMismatch>None</ResolveAssemblyWarnOrErrorOnTargetArchitectureMismatch>
  </PropertyGroup>
```

We also recommend adding the following two XML elements as children of the `<Reference>`
element:

```xml
      <SpecificVersion>False</SpecificVersion>
      <HintPath>..\..\..\Assemblies\x86\Basler.blaze.NET.dll</HintPath>
```

After adding the elements, your project file should contain a segment like this:

```xml
  <PropertyGroup>
    <ResolveAssemblyWarnOrErrorOnTargetArchitectureMismatch>None</ResolveAssemblyWarnOrErrorOnTargetArchitectureMismatch>
  </PropertyGroup>
  <ItemGroup>
    <Reference Include="Basler.blaze.NET, Version=1.0.0.0, Culture=neutral,
               PublicKeyToken=def144ed1a820fcc, processorArchitecture=x86" >
      <SpecificVersion>False</SpecificVersion>
      <HintPath>..\..\..\Assemblies\x86\Basler.blaze.NET.dll</HintPath>
    </Reference>
      <Reference Include="System" />
    <Reference Include="System.Core" />
    <Reference Include="Microsoft.CSharp" />
  </ItemGroup>
```

When finished editing, reload your project by selecting the `Reload project`
entry from the project's context menu. When asked, confirm to close the project
file and to save pending changes.

*Note on the assemblies' processor architecture*

It doesn't matter whether you choose the assembly targeting 32 bit or the one
targeting 64 bit platforms. Since the assemblies have been installed in the
Global Assembly Cache (GAC), the .NET assembly loader will automatically choose
the one matching the architecture your application is built for.

## Python

For operating Basler blaze-101 cameras with python, Basler recommends using
the python `Harvester` module.

Information about the Harvester module can be found here:
<https://github.com/genicam/harvesters>
For information about how to install Harvester, refer to
<https://github.com/genicam/harvesters#installing-harvester-core>.

The following code snippet illustrates how to create a Harvester's image
acquirer for a blaze-101 camera:

```python
import os
import platform
from harvesters.core import Harvester

h = Harvester()

## Location of the Basler blaze GenTL producer.
if platform.system() == "Windows":
    path_to_blaze_cti = os.environ['BASLER_BLAZE_BIN64'] + "\\ProducerBlaze.cti"
elif platform.system() == "Linux":
    path_to_blaze_cti = "/opt/Basler_blaze/lib/gentlproducer/gtl/ProducerBlaze.cti"
else:
    print(f"{platform.system()} is not supported")
    assert False

## Add producer to Harvester.
assert os.path.exists(path_to_blaze_cti)
h.add_file(path_to_blaze_cti)

## Update device list.
h.update()
if h.device_info_list:
    ia = h.create_image_acquirer(0)
else:
    print("No cameras found")
    exit()
```

## Controlling the Heartbeat Timeout

GigE Vision cameras like the blaze-101 camera require the application to 
periodically access the camera by sending special network packets to the camera
in defined intervals. If the camera doesn't receive these heartbeats, it will
consider the connection as broken and won't accept any commands from the
application.

When you run your application, the GenTL producer for ToF cameras will normally
generate these heartbeats. When you set a breakpoint in your application and the
breakpoint is hit, the debugger will suspend all threads including the one
sending the heartbeats. Therefore, when you debug your application and single-
step through your code, no heartbeats are sent to the camera. As a result, the
camera notices a heartbeat timeout and closes the connection. The default value
for the heartbeat timeout is 3000 ms.

To work around this, you have to extend the heartbeat timeout during 
development. You can do this by setting an environment variable named
`GEV_HEARTBEAT_TIMEOUT` to the desired timeout in milliseconds.

For convenience, the GenTL producer automatically sets the heartbeat timeout to
5 minutes when running under a Visual Studio debugger. This behavior can be
overridden by setting the `GEV_HEARTBEAT_TIMEOUT` environment variable as
explained above.

When you extend the heartbeat to a high value and stop your application without
closing the device properly by calling the Close function, you won't be able to
open the camera again and will receive an error stating the device is currently
in use. This can happen if you stop your application using the debugger or if
your application terminates unexpectedly. To open the camera again you must
either wait until the timeout has elapsed or disconnect the network cable from
the camera.
