/*
    This sample illustrates how to get and set camera parameter values.

    For camera configuration and access to other parameters, the GenICam GenApi
    library is used. The GenICam GenApi specification
    (http://www.emva.org/standards-technology/genicam/) defines a format for
    camera description files.
    These files describe the configuration interface of GenICam-compliant
    cameras. The description files are written in XML and describe camera
    registers, their interdependencies, and all other information required for
    accessing high-level features such as Exposure Time, Confidence Threshold,
    or Pixel Format by means of low-level register read and write operations.

    The GenApi library represents the elements of a camera description file
    as software objects called "nodes". For example, a node can represent a
    single camera register, a camera parameter such as Confidence Threshold, a
    set of available parameter values, etc. Each node implements the
    GenApi::INode interface.

    The nodes are linked by different relationships as explained in the
    GenApi standard document available at www.GenICam.org. The complete set of
    nodes is stored in a data structure called a "node map".
    At runtime, a node map is instantiated from an XML description file
    downloaded from the camera.

    The names and types of the parameter nodes can be found by using the blaze
    Viewer tool.

    The CBlazeCamera::GetParameter() functions are used for retrieving smart
    pointers providing access to the parameter nodes.

    In most cases, when you just want to read a camera parameter, set a
    camera parameter or execute a command, you can use the

    * CBlazeCamera::GetCameraParameterValue()
    * CBlazeCamera::SetCameraParameterValue()
    * CBlazeCamera::ExcetuteCommand()

    convenience functions.

    The GenICam GenApi standard document can be downloaded from
    http://www.emva.org/wp-content/uploads/GenICam_Standard_v2_0.pdf
*/


#include <ConsumerImplHelper/BlazeCamera.h>
#include <string>
#include <iostream>

using namespace GenTLConsumerImplHelper;
using namespace GenApi;


// Adjust a value to make it comply with range and increment constraints.
//
// The parameter's minimum and maximum are always considered valid values.
// If the increment is larger than one, the returned value will be: min + (n * inc).
// If the value doesn't meet these criteria, it will be rounded down to ensure compliance.
int64_t Adjust(int64_t val, const CIntegerPtr ptrInt)
{
    const int64_t minimum = ptrInt->GetMin();
    const int64_t maximum = ptrInt->GetMax();
    const int64_t inc = ptrInt->GetInc();

    // Check the lower bound.
    if (val < minimum)
    {
        return minimum;
    }

    // Check the upper bound.
    if (val > maximum)
    {
        return maximum;
    }

    // Check the increment.
    if (inc == 1)
    {
        // Special case: all values are valid.
        return val;
    }
    else
    {
        // The value must be min + (n * inc).
        // Due to the integer division, the value will be rounded down.
        return minimum + (((val - minimum) / inc) * inc);
    }
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_FAILURE;

    CBlazeCamera camera;
    try
    {
        CBlazeCamera::InitProducer();
        camera.OpenFirstCamera();

        ////////////////////////////////////////////////////////////////////
        //
        // How to use the the convenience functions provided by the 
        // CBlazeCamera class for accessing camera parameters
        //
        ////////////////////////////////////////////////////////////////////

        // If you are not interested in a parameter's properties but want to 
        // get or set a value only, you can skip the step of getting a GenApi
        // smart pointer for accessing the parameter value.
        // Instead, use the CBlazeCamera::SetCameraParameterValue() and 
        // CBlazeCamera::GetCameraParameterValue() functions. 
        // SetCameraParameterValue() will clip the value when required, as
        // it also adjusts the value according to increment restrictions.

        // Accessing integer parameters.
        auto threshold = camera.GetCameraParameterValue<int64_t>("ConfidenceThreshold");
        std::cout << "ConfidenceThreshold: " << threshold << std::endl;
        camera.SetCameraParameterValue("ConfidenceThreshold", threshold);

        // Accessing boolean parameters.
        auto isTemporalFilterEnabled = camera.GetCameraParameterValue<bool>("TemporalFilter");
        std::cout << "Temporal filter is " << (isTemporalFilterEnabled ? "enabled" : "disabled")
            << std::endl;

        // Accessing enumeration and string parameters.
        camera.SetCameraParameterValue("ComponentSelector", "Range");
        camera.SetCameraParameterValue("PixelFormat", "Coord3D_C16");
        std::cout << "Current pixel format for component "
            << camera.GetCameraParameterValue<std::string>("ComponentSelector")
            << " is " << camera.GetCameraParameterValue<std::string>("PixelFormat")
            << std::endl;

        // Accessing floating point parameters
        auto exposureTime = camera.GetCameraParameterValue<double>("ExposureTime");
        camera.SetCameraParameterValue("ExposureTime", exposureTime);

        // Trying to access a parameter using the wrong type or trying to 
        // set a read-only parameter will result in GenICam::GenericExceptions
        // to be thrown.
        try
        {
            // ConfidenceThreshild is an integer, not a bool.
            auto val = camera.GetCameraParameterValue<bool>("ConfidenceThreshold");
        }
        catch (const GenICam::GenericException& e)
        {
            std::cout << e.GetDescription() << std::endl;
        }
        try
        {
            // SensorWidth is read-only
            camera.SetCameraParameterValue("SensorWidth", 1024);
        }
        catch (const GenICam::GenericException& e)
        {
            std::cout << e.GetDescription() << std::endl;
        }

        // Issue a command. Example: before reading the camera's current 
        // time stamp, it must be latched first.
        camera.ExecuteCameraCommand("TimestampLatch");
        std::cout << "Current timestamp: "
            << camera.GetCameraParameterValue<int64_t>("TimestampLatchValue")
            << std::endl;


        ////////////////////////////////////////////////////////////////////
        //
        // How to use the classes and functions provided by the GenApi
        // library for accessing camera parameters and retrieving more information
        // about camera parameter properties.
        //
        ////////////////////////////////////////////////////////////////////

        // 
        // Common properties of all parameter types
        //

        // We use GenApi's smart pointer like classes for a convenient way to get access 
        // to parameters. GenApi supports different parameter types, for each of them 
        // there is a corresponding parameter class: CIntgegerPtr, CFloatPtr, CEnumerationPtr, 
        // CBooleanPtr, etc.

        CIntegerPtr ptrFancyParameter = camera.GetParameter("FancyParameter");

        // Although the camera doesn't provide this fancy parameter, the camera class
        // returns a smart pointer. Accessing the smart pointer will throw an exception 
        // since there is no corresponding GenApi node object the pointer could provide 
        // access to. Trying to access the value or the properties of ptrFancyParameter will result
        // in an GenICam::GenericException. 
        // Check to see whether GetParameter returned a valid pointer to a parameter:
        if (!ptrFancyParameter.IsValid())
        {
            std::cout << "Camera doesn't provide the parameter 'FancyParameter'" << std::endl;
        }

        // Parameters might be read-only. When accessing read-only parameters, we will get
        // an exception. The GenApi::IsWritable() function can be used to check whether
        // a value can be set.
        // GenApi provides some further convenience functions for checking the accessibility
        // of nodes: GenApi::IsReadable(), GenApi::IsImplemended().
        // These functions apply to all parameter types.

        // Of course, a non-existent parameter is not readable...
        if (!IsReadable(ptrFancyParameter))
        {
            std::cout << "'FancyParameter' is not readable." << std::endl;
        }

        // Measured temperatures are not writable...
        CFloatPtr ptrDeviceTemperature = camera.GetParameter("DeviceTemperature");
        if (!IsWritable(ptrDeviceTemperature))
        {
            std::cout << "Device temperature is not writable." << std::endl;
        }
        if (IsReadable(ptrDeviceTemperature))
        {
            std::cout << "Current temperature is " << ptrDeviceTemperature->GetValue() << " degrees Celsius" << std::endl;
        }

        // As a more convenient alternative to the CBlazeCamera::GetParameter() 
        // function, you can use the CBlazeCamera::GetParameter<T>() template
        // function for retrieving a GenApi smart pointer.
        // This function checks whether there is a parameter with the name that
        // is requested, if it is of the requested type, and if it is accessible.
        // When the checks fail, a GenICam::GenericException is thrown.
        // By default, the function checks for read and write access. 
        // Read-only or write-only access can be requested by specifying 
        // GenApi::RO resp. GenApi::RW as additional argument.

        // Happy path.
        auto ptrDepthMax = camera.GetParameter<GenApi::CIntegerPtr>("DepthMax");
        std::cout << "DepthMax: current value: " << ptrDepthMax->GetValue() << "mm. ";
        std::cout << "Value range: ptrDepthMax [" << ptrDepthMax->GetMin()
            << ".." << ptrDepthMax->GetMax() << "]" << std::endl;

        // Requesting read-write access for a read-only parameter will throw
        // an exception.
        try
        {
            auto ptrSensorWidth = camera.GetParameter<GenApi::CIntegerPtr>("SensorWidth");
        }
        catch (const GenICam::GenericException& e)
        {
            std::cout << e.GetDescription() << std::endl;
        }

        // This will work, we request read-only access by passing GenApi::RO as 
        // desired access mode.
        auto ptrSensorWidth = camera.GetParameter<GenApi::CIntegerPtr>("SensorWidth", GenApi::RO);
        std::cout << "Sensor width = " << ptrSensorWidth->GetValue() << " px" << std::endl;

        // Trying to retrieve a smart pointer of the wrong type will throw an exception.
        try
        {
            // DepthMax is an integer parameter, not a float.
            auto ptrWrongType = camera.GetParameter<GenApi::CFloatPtr>("DepthMax");
        }
        catch (const GenICam::GenericException& e)
        {
            std::cout << e.GetDescription() << std::endl;
        }

        //
        // Usage of integer parameters
        //

        // Some parameters may have restrictions, i.e., the values must be a multiple of a certain
        // increment. Use GetInc/GetMin/GetMax to make sure you set a valid value, as shown
        // in the Adjust() function above.
        auto ptrConfidenceThreshold = camera.GetParameter<GenApi::CIntegerPtr>("ConfidenceThreshold");
        // As this is for demonstration purposes only, we don't change the value.
        int64_t newThresh = ptrConfidenceThreshold->GetValue();
        newThresh = Adjust(newThresh, ptrConfidenceThreshold);
        ptrConfidenceThreshold->SetValue(newThresh);

        // Alternatively, you can use the SetParameterValue() convenience function 
        // that adjusts the value automatically.
        SetParameterValue(ptrConfidenceThreshold, newThresh);

        // Values of all parameter types can be converted to a string using the ToString() method.
        std::cout << "Confidence Threshold : " << ptrConfidenceThreshold->ToString() << std::endl;

        // 
        // Usage of enumeration parameters, selectors, and Boolean parameters
        //
        // Enumeration parameters, as the PixelFormat, represent a defined set of named values.
        // Selectors are enumeration parameters influencing the state of the parameters
        // that are "selected by" the selector.
        // The ComponentSelector feature, as an example, is a selector. The feature controls 
        // the number of parts an acquired image consists of. Each part can be individually 
        // enabled by using the ComponentEnable parameter.
        // For some parts there are multiple options for the part's pixel format.
        // Thus, the ComponentSelector "selects" the PixelFormat and the Component parameters.
        //
        // All parameter types provide a corresponding FromString() method allowing to set
        // a parameter by using a string. The FromString() method is the only way to set the
        // value of an enumeration parameter.
        // 
        // Example: Iterate over all parts and enable those parts that support the Mono16
        // pixel format.

        auto ptrComponentSelector = camera.GetParameter<GenApi::CEnumerationPtr>("ComponentSelector");
        auto ptrComponentEnable = camera.GetParameter<GenApi::CBooleanPtr>("ComponentEnable");
        auto ptrPixelFormat(camera.GetParameter<GenApi::CEnumerationPtr>("PixelFormat"));
        NodeList_t entries;
        ptrComponentSelector->GetEntries(entries);
        for (NodeList_t::const_iterator entry = entries.begin(); entry != entries.end(); ++entry)
        {
            CEnumEntryPtr ptrEntry(*entry);
            std::cout << ptrEntry->GetSymbolic() << " part is " << (IsAvailable(ptrEntry) ? "supported" : "not supported") << std::endl;
            if (IsAvailable(ptrEntry))
            {
                // First, select the part to configure...
                ptrComponentSelector->FromString(ptrEntry->GetSymbolic());
                // ... then access the parameters for the part.
                if (IsAvailable(ptrPixelFormat->GetEntryByName("Mono16")))
                {
                    ptrPixelFormat->FromString("Mono16");
                    ptrComponentEnable->SetValue(true);
                }
                std::cout << "Pixel format of part " << ptrComponentSelector->ToString() << ": " << ptrPixelFormat->ToString() << std::endl;
                std::cout << "Part is " << (ptrComponentEnable->GetValue() ? "enabled" : "disabled") << std::endl;
            }
        }

        //
        // Usage of floating point parameters
        //
        // Configuring the camera's exposure time
        // Note: Generally, you should not adjust the exposure time. Unless you
        // don't notice saturation effects or missing depth information due
        // to close objects being overexposed, you should stick to the default
        // exposure time value. 
        //
        auto ptrExposureTime(camera.GetParameter<GenApi::CFloatPtr>("ExposureTime"));
        // Retrieve the current exposure time.
        double currentExposureTime = ptrExposureTime->GetValue();
        // For demonstration purposes, we write back the current value.
        ptrExposureTime->SetValue(currentExposureTime);
        std::cout << "Set exposure time to " << ptrExposureTime->GetValue() << std::endl;
        std::cout << "Value range of exposure time: Min= " << ptrExposureTime->GetMax() <<
            " Max= " << ptrExposureTime->GetMax() << std::endl;

        camera.Close();
        exitCode = EXIT_SUCCESS;
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used 
        // to check if the device is still connected.
        if (camera.IsOpen() && !camera.IsConnected())
        {
            std::cerr << "Camera has been removed." << std::endl;
        }
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}

