/* 
This sample demonstrates how to grab images from both a Basler blaze and a
Basler GigE color camera and how to access the image and depth data.
The cameras are triggered using software triggers.

For accessing the color camera, Basler pylon version 6 or higher with GigE 
camera support and the GigE GenTL producer must be installed.

*/
#include <ConsumerImplHelper/BlazeCamera.h>
#include <ConsumerImplHelper/GEVCamera.h>
#include <opencv2/opencv.hpp>

using namespace GenTLConsumerImplHelper;

const std::map<std::string, cv::ColorConversionCodes> BayerFormat = {
    { "BayerGR8", cv::COLOR_BayerGR2RGB },
    { "BayerRG8", cv::COLOR_BayerRG2RGB },
    { "BayerBG8", cv::COLOR_BayerBG2RGB },
    { "BayerGB8", cv::COLOR_BayerGB2RGB }
};

class Sample
{
public:
    int run();

private:
    void setupToFCamera();
    void setupColorCamera();
    void processToFData(const GrabResult& result);
    void processColorData(const GrabResult& result);

private:
    CBlazeCamera    m_blazeCamera;
    std::string     m_bayerPattern; // The camera's color pattern.
    CGEVCamera      m_colorCamera;
};

void Sample::setupToFCamera()
{
    m_blazeCamera.OpenFirstCamera();

    //If there are multiple cameras connected and you want to open a specific one, use
    //    the CBlazeCamera::Open(CameraInfoKey, string) method.
    std::cout << "Connected to blaze camera " << m_blazeCamera.GetCameraInfo().strDisplayName << std::endl;

    // Enable depth map.
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Range");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", true);
    m_blazeCamera.SetCameraParameterValue("PixelFormat", "Coord3D_C16");

    // Enable the intensity image.
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Intensity");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", true);

    // Enable the confidence map.
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Confidence");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", true);

    // Configure the camera for software trigger.
    m_blazeCamera.SetCameraParameterValue("TriggerMode", "On");
    m_blazeCamera.SetCameraParameterValue("TriggerSource", "Software");

    // Prepare windows for displaying intensity image and depth map.
    cv::namedWindow("Intensity", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("Depth", cv::WINDOW_AUTOSIZE);
}

void Sample::setupColorCamera()
{
    CameraList cameras = CGEVCamera::EnumerateCameras();
    std::string serialnumber;
    // Since the blaze camera is a GigE Vision camera, it will be enumerated too.
    // We iterate through all found cameras and open the first one which is not 
    // a blaze camera.
    for (auto cam : cameras) {
        if (cam.strModelName.find("blaze") == std::string::npos) {
            serialnumber = cam.strSerialNumber;
            break;
        }
    }
    if (serialnumber.empty())
    {
        throw std::runtime_error("No GigE camera found!");
    }

    m_colorCamera.Open(SerialNumber, serialnumber);
    std::cout << "Connected to GEV camera " << m_colorCamera.GetCameraInfo().strDisplayName << std::endl;

    auto ptrPixelFormat = m_colorCamera.GetParameter<GenApi::CEnumerationPtr>("PixelFormat");

    // Figure out which 8-bit Bayer pixel format the camera supports.
    // If the camera supports an 8-bit Bayer format, enable the format. Otherwise, throw an
    // exception.
    m_bayerPattern = "";
    GenApi::CEnumEntryPtr ptrFormatEntry;
    for (auto const& formatName : BayerFormat) {
        ptrFormatEntry = ptrPixelFormat->GetEntryByName(formatName.first.c_str());
        if (GenApi::IsAvailable(ptrFormatEntry)) {
            m_bayerPattern = formatName.first;
            break;
        }
    }

    if (m_bayerPattern != "") {
        ptrPixelFormat->FromString(ptrFormatEntry->GetSymbolic());
    } else {
        throw RUNTIME_EXCEPTION("Camera doesn't support an 8-bit Bayer pixel format.");
    }

    // Configure the camera for software trigger.
    // Each software trigger will start the acquisition of one single frame.
    m_colorCamera.SetCameraParameterValue("AcquisitionMode", "Continuous");
    m_colorCamera.SetCameraParameterValue("TriggerSelector", "FrameStart");
    m_colorCamera.SetCameraParameterValue("TriggerMode", "On");
    m_colorCamera.SetCameraParameterValue("TriggerSource", "Software");

    // Prepare windows for displaying intensity image and depth map.
    cv::namedWindow("Color", cv::WINDOW_AUTOSIZE);
}

void Sample::processToFData(const GrabResult& result)
{
    BufferParts parts;
    m_blazeCamera.GetBufferParts(result, parts);

    const int width = (int)parts[0].width;
    const int height = (int)parts[0].height;
    const int count = width * height;
    cv::Mat intensity = cv::Mat(height, width, CV_16UC1, parts[1].pData);
    cv::Mat depth = cv::Mat(height, width, CV_16UC1, parts[0].pData);

    // -------   Display the images
    cv::imshow("Intensity", intensity);
    cv::imshow("Depth", depth);
}

void Sample::processColorData(const GrabResult& result)
{
    BufferParts parts;
    m_colorCamera.GetBufferParts(result, parts);

    // The camera delivers the color information as a Bayer-coded raw image.
    // Before accessing the RGB data, a Bayer decoding must be performed.
    cv::Mat rawImage(int(parts[0].height), int(parts[0].width), CV_8UC1, parts[0].pData);
    cv::Mat color;
    cv::cvtColor(rawImage, color, BayerFormat.at(m_bayerPattern));

    // -------   Display the images
    cv::imshow("Color", color);
}

int Sample::run()
{
    int result = EXIT_SUCCESS;
    const int nBuffers = 2;   // Number of buffers used per camera
    int key = 0;

    try {
        // Open and configure the cameras.
        setupToFCamera();
        setupColorCamera();

        // Allocate memory buffers used for grabbing.
        m_blazeCamera.PrepareAcquisition(nBuffers);
        m_colorCamera.PrepareAcquisition(nBuffers);

        unsigned int idxNextBuffer = 0;

        // For each camera, enqueue one buffer which will receive the image data of the next image
        // the corresponding camera sends.
        m_blazeCamera.QueueBuffer(idxNextBuffer);
        m_colorCamera.QueueBuffer(idxNextBuffer);

        // Start the acquisition on the host. Now, the enqueued buffers will be filled with data sent by the cameras.
        m_blazeCamera.StartAcquisition();
        m_colorCamera.StartAcquisition();

        // The camera won't send any data until they are started.
        m_blazeCamera.IssueAcquisitionStartCommand();
        m_colorCamera.IssueAcquisitionStartCommand();

        // The cameras run in software trigger mode. They will expose and transfer a new image for
        // each software trigger they receive.
        // To optimize bandwidth usage, the color camera is triggered first to
        // allow it to already transfer image data while the blaze camera is still internally
        // processing the acquired raw data.
        m_colorCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the color camera.
        m_blazeCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the blaze camera.

        std::cout << "To exit, press 'q' in one of the image windows" << std::endl;

        // Grab loop. Acquire and process data from both cameras.
        // While the data is being processed for each camera, a new buffer will be filled in
        // the background.
        int nSuccess = 0;
        while (result == EXIT_SUCCESS) { // Process n images from each camera.
            GrabResult ToFResult;
            GrabResult ColorResult;
            // Wait up to 1000 ms until the enqueued buffers have been completely filled.
            m_blazeCamera.GetGrabResult(ToFResult, 1000);
            m_colorCamera.GetGrabResult(ColorResult, 1000);

            // Enqueue the next buffers.
            // The new buffers will be filled asynchronously in the background.
            // Each buffer will be filled with the next image received from the corresponding camera.
            idxNextBuffer++;
            idxNextBuffer %= nBuffers;
            m_blazeCamera.QueueBuffer(idxNextBuffer);
            m_colorCamera.QueueBuffer(idxNextBuffer);
            // Issue software triggers.
            m_colorCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the color camera.
            m_blazeCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the blaze camera.

            // While the next buffers are acquired, process the current ones.
            if (ToFResult.status != GrabResult::Ok) {
                std::cerr << "Failed to grab image from blaze camera." << std::endl;
                if (ToFResult.status == GrabResult::Timeout) {
                    std::cerr << "Timeout occurred." << std::endl;
                }
                result = EXIT_FAILURE;
            }

            if (ColorResult.status != GrabResult::Ok) {
                std::cerr << "Failed to grab image from color camera." << std::endl;
                if (ColorResult.status == GrabResult::Timeout) {
                    std::cerr << "Timeout occurred." << std::endl;
                }
                result = EXIT_FAILURE;
            }

            if (result == EXIT_SUCCESS) {
                nSuccess++;
                // Successfully acquired the images.

                // Now process the data.
                processToFData(ToFResult);
                processColorData(ColorResult);
            }

            key = cv::waitKey(1);
            if ('q' == (char)key) {
                break;
            }
        }

        // Shut down acquisition. Stops cameras and frees all grab-related resources.
        m_blazeCamera.FinishAcquisition();
        m_colorCamera.FinishAcquisition();
    }
    catch (std::runtime_error& e)
    {
        std::cerr << "Exception occurred: " << e.what() << std::endl;
        result = EXIT_FAILURE;
    } catch (const GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        result = EXIT_FAILURE;
    }

    // clean-up
    if (m_colorCamera.IsOpen())
    {
        try
        {
           // Disable software trigger.
            m_colorCamera.SetCameraParameterValue("TriggerMode", "Off");

            m_colorCamera.Close();
        }
        catch(...)
        {}
    }
    if (m_blazeCamera.IsOpen())
    {
        try
        {
            // Disable software trigger.
            m_blazeCamera.SetCameraParameterValue("TriggerMode", "Off");

            m_blazeCamera.Close();
        }
        catch (...)
        {
        }
    }

    return result;
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try {
        CBlazeCamera::InitProducer();
        CGEVCamera::InitProducer();

        exitCode = Sample().run();
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producers and all of their resources.
    // Note: Don't call TerminateProducer() until the destructors of the camera classes
    // have been called. The destructors may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized()) {
        CBlazeCamera::TerminateProducer();
    }

    if (CGEVCamera::IsProducerInitialized()) {
        CGEVCamera::TerminateProducer();
    }

    return exitCode;
}
