/*
This sample demonstrates how to calibrate a system consisting of a Basler GigE color camera
and a Basler blaze camera.

The project consists of two executables: the calibration and the fusion. In the first step,
the calibration executable must be used to determine the relative position of the cameras
to each other.

After the camera system has been calibrated successfully, start the fusion as the second step.
The sample program uses the data of the color camera to display a colored point cloud. This step
requires the 'calibration.xml' file that was created during calibration. This file contains the
intrinsic camera parameters of the blaze camera and the color camera as well as the values for
the relative rotation and translation between the cameras.

To access the color camera, the Basler pylon Camera Software Suite (version 6 or higher) with
GigE camera support and the GigE GenTL Producer must be installed.
*/


#include <ConsumerImplHelper/BlazeCamera.h>
#include <ConsumerImplHelper/GEVCamera.h>

#include <opencv2/calib3d.hpp>
#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>

#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/common/common_headers.h>
#include <vtkRenderWindow.h>


using namespace GenTLConsumerImplHelper;

const std::map<std::string, cv::ColorConversionCodes> BayerFormat = {
    {"BayerGR8", cv::COLOR_BayerGR2BGR},
    {"BayerRG8", cv::COLOR_BayerRG2BGR},
    {"BayerBG8", cv::COLOR_BayerBG2BGR},
    {"BayerGB8", cv::COLOR_BayerGB2BGR}
};

class Fusion
{
public:
    int run();

private:
    void loadCalibrationFile();
    void setupVisualizer();
    void setupBlazeCamera();
    void setupColorCamera();
    cv::Mat processBlazeData(const GrabResult& result);
    cv::Mat processColorData(const GrabResult& result);
    cv::Mat warpColorToDepth(const cv::Mat pointcloud, const cv::Mat& color);

private:
    CBlazeCamera    m_blazeCamera;
    std::string     m_bayerPattern; // The camera's color pattern.
    CGEVCamera      m_colorCamera;

    pcl::visualization::PCLVisualizer::Ptr m_visualizer;

    cv::Mat         m_rotation;
    cv::Mat         m_translation;
    cv::Mat         m_colorCameraMatrix;
    cv::Mat         m_colorDistortion;
};


void Fusion::loadCalibrationFile()
{
    // It is assumed that the 'calibration.xml' contains information about the relative orientation
    // of the cameras to each other and the optical calibration of the color camera.
    // The calibration program can be used to create the file.
    const std::string filename = std::string("calibration_") + m_blazeCamera.GetCameraInfo().strSerialNumber + "_"
                                 + m_colorCamera.GetCameraInfo().strSerialNumber + ".xml";
    const std::string path = std::string(CALIBRATION_DIR) + filename;
    std::cout << "Loading the calibration file: " << path << std::endl;
    cv::FileStorage fs(path, cv::FileStorage::READ);

    if (!fs.isOpened()) {
        std::string msg("For the fusion of the data, calibration must first be performed.\nMake sure that there is a valid calibration with file name 'calibration.xml' in the directory.");
        throw RUNTIME_EXCEPTION(msg);
    }

    fs["rotation"] >> m_rotation;
    fs["translation"] >> m_translation;
    fs["colorCameraMatrix"] >> m_colorCameraMatrix;
    fs["colorDistortion"] >> m_colorDistortion;
}

void Fusion::setupVisualizer()
{
    // Set up a PCLVisualizer, i.e. background color, coordinate system and initial camera position
    m_visualizer = pcl::visualization::PCLVisualizer::Ptr(new pcl::visualization::PCLVisualizer("Basler AG - RGBD-Fusion"));
    m_visualizer->setBackgroundColor(0, 0, 0);
    m_visualizer->setShowFPS(false);
    m_visualizer->addCoordinateSystem(0.1);
    m_visualizer->setCameraPosition(0.0, 0.1, -2.5, 0, -1.0, 0);
    m_visualizer->getRenderWindow()->GlobalWarningDisplayOff();
    vtkObject::GlobalWarningDisplayOff();
}

void Fusion::setupBlazeCamera()
{
    m_blazeCamera.Open(ModelName, "blaze-101");

    // If there are multiple cameras connected and you want to open a specific one, use
    // the CToFCamera::Open(CameraInfoKey, string) method.
    std::cout << "Connected to ToF camera " << m_blazeCamera.GetCameraInfo().strDisplayName << std::endl;

    // Enable range data
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Range");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", true);
    m_blazeCamera.SetCameraParameterValue("PixelFormat", "Coord3D_ABC32f");
    m_blazeCamera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateA");
    m_blazeCamera.SetCameraParameterValue("Scan3dInvalidDataValue", 0.0);
    m_blazeCamera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateB");
    m_blazeCamera.SetCameraParameterValue("Scan3dInvalidDataValue", 0.0);
    m_blazeCamera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateC");
    m_blazeCamera.SetCameraParameterValue("Scan3dInvalidDataValue", 0.0);

    // Enable the intensity image.
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Intensity");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", true);
    m_blazeCamera.SetCameraParameterValue("PixelFormat", "Mono16");

    // Disable the confidence map.
    m_blazeCamera.SetCameraParameterValue("ComponentSelector", "Confidence");
    m_blazeCamera.SetCameraParameterValue("ComponentEnable", false);

    // Exposure time
    m_blazeCamera.SetCameraParameterValue("ExposureTime", 1000.0);

    // Configure the camera for software trigger.
    m_blazeCamera.SetCameraParameterValue("TriggerMode", "On");
    m_blazeCamera.SetCameraParameterValue("TriggerSource", "Software");
}

void Fusion::setupColorCamera()
{
    CameraList cameras = CGEVCamera::EnumerateCameras();
    std::string serialnumber = "";
    // Since the blaze camera is a GigE Vision camera, it will be enumerated too.
    // We iterate through all found cameras and open the first one which is not
    // a blaze camera.
    for (auto cam : cameras) {
        if (cam.strModelName.find("blaze") == std::string::npos) {
            serialnumber = cam.strSerialNumber;
            break;
        }
    }
    if (serialnumber.empty()) {
        throw RUNTIME_EXCEPTION("No 2D camera found.");
    }
    m_colorCamera.Open(SerialNumber, serialnumber);
    std::cout << "Connected to color camera " << m_colorCamera.GetCameraInfo().strDisplayName << std::endl;

    GenApi::CEnumerationPtr ptrPixelFormat = m_colorCamera.GetParameter("PixelFormat");

    // Figure out which 8-bit Bayer pixel format the camera supports.
    // If the camera supports an 8-bit Bayer format, enable the format. Otherwise, throw an
    // exception.
    m_bayerPattern = "";
    GenApi::CEnumEntryPtr ptrFormatEntry;
    for (auto const& formatName : BayerFormat) {
        ptrFormatEntry = ptrPixelFormat->GetEntryByName(formatName.first.c_str());
        if (GenApi::IsAvailable(ptrFormatEntry)) {
            m_bayerPattern = formatName.first;
            break;
        }
    }
    if (m_bayerPattern != "") {
        ptrPixelFormat->FromString(ptrFormatEntry->GetSymbolic());
    } else {
        throw RUNTIME_EXCEPTION("Camera doesn't support an 8-bit Bayer pixel format.");
    }

    // Setting the exposure time
    m_colorCamera.SetCameraParameterValue("ExposureAuto", "Off");
    int64_t sfncVersion = 1;
    try {
        sfncVersion = m_colorCamera.GetCameraParameterValue<int64_t>("DeviceSFNCVersionMajor");
    } catch (const GenICam::GenericException&) {
        std::cerr << "DeviceSFNCVersionMajor could not be read. Major version 1 is assumed." << std::endl;
    }
    const float exposureTime = 15000.0f;
    if (sfncVersion < 2) {
        // ace classic/U/L GigE Cameras
        m_colorCamera.SetCameraParameterValue("ExposureTimeAbs", exposureTime);
    } else {
        // ace 2, boost, and dart IMX Cameras
        m_colorCamera.SetCameraParameterValue("ExposureTime", exposureTime);
    }

    // If your network hardware can't handle the incoming packet rate, it is
    // useful to increase the delay between packet transmissions.
    m_colorCamera.SetCameraParameterValue("GevSCPD", 10000);

    // Increasing the packet size will reduce network load. Packet sizes
    // greater than 1500 require so called Jumbo Frames to be activated for
    // your network interface adapter. We recommend to enable Jumbo Frames in
    // your network adatapter's setting and to uncomment the following line.
    //m_colorCamera.SetCameraParameterValue("GevSCPSPacketSize", 6000);

    // White balance is adjusted continuously while images are being acquired.
    m_colorCamera.SetCameraParameterValue("BalanceWhiteAuto", "Continuous");

    // Configure the camera for software trigger.
    // Each software trigger will start the acquisition of one single frame.
    m_colorCamera.SetCameraParameterValue("AcquisitionMode", "Continuous");
    m_colorCamera.SetCameraParameterValue("TriggerSelector", "FrameStart");
    m_colorCamera.SetCameraParameterValue("TriggerMode", "On");
    m_colorCamera.SetCameraParameterValue("TriggerSource", "Software");
}

cv::Mat Fusion::processBlazeData(const GrabResult& result)
{
    BufferParts parts;
    m_blazeCamera.GetBufferParts(result, parts);

    const int width = (int)parts[0].width;
    const int height = (int)parts[0].height;
    cv::Mat pointcloud = cv::Mat(height, width, CV_32FC3, parts[0].pData);

    return pointcloud;
}

cv::Mat Fusion::processColorData(const GrabResult& result)
{
    BufferParts parts;
    m_colorCamera.GetBufferParts(result, parts);

    // The camera delivers the color information as a Bayer-coded raw image.
    // Before accessing the RGB data, a Bayer decoding must be performed.
    cv::Mat rawImage(int(parts[0].height), int(parts[0].width), CV_8UC1, parts[0].pData);
    cv::Mat color;
    cv::cvtColor(rawImage, color, BayerFormat.at(m_bayerPattern));

    return color;
}

cv::Mat Fusion::warpColorToDepth(const cv::Mat pointcloud, const cv::Mat& color)
{
    const int width = pointcloud.cols;
    const int height = pointcloud.rows;
    cv::Mat coloredDepth(pointcloud.size(), CV_8UC3);

    // Project the 3d point into the color camera
    cv::Mat pointcloudVec = pointcloud.reshape(3, 1);
    cv::Mat projectedPoints;
    cv::projectPoints(pointcloudVec, m_rotation, m_rotation* m_translation, m_colorCameraMatrix, m_colorDistortion, projectedPoints);

    const cv::Vec3b invalidPoint(0, 0, 0);
    cv::Rect colorRect(cv::Point(), color.size());
    for (int row = 0; row < height; ++row) {
        for (int col = 0; col < width; ++col) {
            const int idx = row * width + col;
            const cv::Vec3f& point3d = pointcloudVec.at<cv::Vec3f>(idx);
            if (point3d[2] != 0.0f) {
                const cv::Point2f& colorPoint = projectedPoints.at<cv::Point2f>(idx);
                if (colorRect.contains(colorPoint)) {
                    coloredDepth.at<cv::Vec3b>(row, col) = color.at<cv::Vec3b>(colorPoint);
                } else {
                    // No color information avalable
                    coloredDepth.at<cv::Vec3b>(row, col) = invalidPoint;
                }
            } else {
                // No depth information available for this pixel. Zero it.
                coloredDepth.at<cv::Vec3b>(row, col) = invalidPoint;
            }
        }
    }

    return coloredDepth;
}

int Fusion::run()
{
    int result = EXIT_SUCCESS;
    const int nBuffers = 3;   // Number of buffers used per camera
    int key = 0;

    try {
        // Setup the pcl visualizer
        std::string cloudID = "colored_cloud";
        pcl::PointCloud<pcl::PointXYZRGB>::Ptr pointcloud(new pcl::PointCloud<pcl::PointXYZRGB>);
        setupVisualizer();
        m_visualizer->addPointCloud(pointcloud, cloudID);
        m_visualizer->setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 2, cloudID);

        // Open and configure the cameras.
        setupBlazeCamera();
        setupColorCamera();

        // Load camera calibration
        loadCalibrationFile();

        // Allocate memory buffers used for grabbing.
        m_blazeCamera.PrepareAcquisition(nBuffers);
        m_colorCamera.PrepareAcquisition(nBuffers);

        // For each camera, enqueue all buffers to be filled with image data.
        for (size_t i = 0; i < nBuffers; ++i) {
            m_blazeCamera.QueueBuffer(i);
            m_colorCamera.QueueBuffer(i);
        }

        // Start the acquisition on the host. Now, the enqueued buffers will be filled with data sent by the cameras.
        m_blazeCamera.StartAcquisition();
        m_colorCamera.StartAcquisition();

        // The camera won't send any data until they are started.
        m_blazeCamera.IssueAcquisitionStartCommand();
        m_colorCamera.IssueAcquisitionStartCommand();
        m_colorCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the color camera.
        m_blazeCamera.ExecuteCameraCommand("TriggerSoftware"); // Issue software trigger for the blaze camera.

        // The cameras run in software trigger mode. They will expose and transfer a new image for
        // each software trigger they receive.

        std::cout << std::endl << "Fusion of color and depth data" << std::endl;
        std::cout << "  - Press 'q' in the viewer to exit" << std::endl;
        std::cout << std::endl;

        // Grab loop. Acquire and process data from both cameras.
        // While the data is being processed for each camera, a new buffer will be filled in
        // the background.
        while (!m_visualizer->wasStopped()) { // Process n images from each camera.
            GrabResult blazeResult;
            GrabResult colorResult;
            // Wait up to 1000 ms until the enqueued buffers have been completely filled.
            m_blazeCamera.GetGrabResult(blazeResult, 1000);
            m_colorCamera.GetGrabResult(colorResult, 1000);

            // Issue software triggers.
            m_colorCamera.ExecuteCameraCommand("TriggerSoftware");
            m_blazeCamera.ExecuteCameraCommand("TriggerSoftware");

            // While the next buffers are acquired, process the current ones.
            if (blazeResult.status != GrabResult::Ok || colorResult.status != GrabResult::Ok) {
                std::cerr << "Failed to grab image." << std::endl;

                if (blazeResult.status != GrabResult::Timeout) {
                    // Put the buffer back into the acquisition engine's input queue.
                    m_blazeCamera.QueueBuffer(blazeResult.hBuffer);
                }

                if (colorResult.status != GrabResult::Timeout) {
                    // Put the buffer back into the acquisition engine's input queue.
                    m_colorCamera.QueueBuffer(colorResult.hBuffer);
                }

                // In case of error the image pair is skipped
                continue;
            }

            // Process the data.
            cv::Mat blazeImg = processBlazeData(blazeResult);
            cv::Mat colorImg = processColorData(colorResult);
            cv::Mat coloredDepth = warpColorToDepth(blazeImg, colorImg);

            // Generation of a coloured point cloud from the colour image
            // warped into the depth image and the point cloud.
            const int width = blazeImg.cols;
            const int height = blazeImg.rows;
            pointcloud = pcl::PointCloud<pcl::PointXYZRGB>::Ptr(new pcl::PointCloud<pcl::PointXYZRGB>(width, height));
            for (int row = 0; row < height; row += 1) {
                for (int col = 0; col < width; col += 1) {
                    auto& coloredPoint = pointcloud->at(col, row); // PCL uses column/row instead of row/column

                    const cv::Vec3f point3d = blazeImg.at<cv::Vec3f>(row, col) / 1000.0f; // Scaling from mm to m for PCL
                    coloredPoint.x = point3d[0];
                    coloredPoint.y = point3d[1];
                    coloredPoint.z = point3d[2];

                    const cv::Vec3b color = coloredDepth.at<cv::Vec3b>(row, col);
                    coloredPoint.r = color[0];
                    coloredPoint.g = color[1];
                    coloredPoint.b = color[2];
                }
            }

            // Update the pointcloud and refresh the visualizer
            m_visualizer->updatePointCloud(pointcloud, cloudID);
            m_visualizer->spinOnce();

            // We finished processing the data, put the buffer back into the acquisition
            // engine's input queue to be filled with new image data.
            m_blazeCamera.QueueBuffer(blazeResult.hBuffer);
            m_colorCamera.QueueBuffer(colorResult.hBuffer);
        }

        // Shut down acquisition. Stops cameras and frees all grab-related resources.
        m_blazeCamera.FinishAcquisition();
        m_colorCamera.FinishAcquisition();

    } catch (const GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        result = EXIT_FAILURE;
    }
    // clean-up
    if (m_colorCamera.IsOpen())
    {
        try
        {
            // Disable software trigger.
            m_colorCamera.SetCameraParameterValue("TriggerMode", "Off");

            m_colorCamera.Close();
        }
        catch (...)
        {
        }
    }
    if (m_blazeCamera.IsOpen())
    {
        try
        {
            // Disable software trigger.
            m_blazeCamera.SetCameraParameterValue("TriggerMode", "Off");

            m_blazeCamera.Close();
        }
        catch (...)
        {
        }
    }

    return result;
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try {
        CBlazeCamera::InitProducer();
        CGEVCamera::InitProducer();

        Fusion sample;
        exitCode = sample.run();
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producers and all of their resources.
    // Note: Don't call TerminateProducer() until the destructors cameras of the camera classes
    // have been called. The destructors may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized()) {
        CBlazeCamera::TerminateProducer();
    }

    if (CGEVCamera::IsProducerInitialized()) {
        CGEVCamera::TerminateProducer();
    }

    return exitCode;
}
