/*
   This sample illustrates how to extract depth maps from point clouds.
   Extracting a depth map from a point cloud is required when you want to process point clouds
   and depth maps at the same time. 
   Since you can acquire depth information either as point clouds or as depth maps only, 
   the application has to extract depth maps from acquired point clouds.

   When your application requires depth maps only, refer to the 'GrabRangeMap' sample.

   The sample uses OpenCV only for displaying the results. The depth map extraction doesn't
   depend on OpenCV.

*/

#include <ConsumerImplHelper/BlazeCamera.h>
#include <opencv2/opencv.hpp>

using namespace GenTLConsumerImplHelper;


// Checks whether a given 3d point represents a valid coordinate.
// The Scan3dInvalidDataValue is used to identify a non-valid pixel.
bool isValid(const CBlazeCamera::Coord3D* point, float invalidDataValue)
{
    return std::isnan(invalidDataValue) ? !std::isnan(point->z) : point->z != invalidDataValue;
}

// Calculates a grayscale depth map from point cloud data sent from a Basler blaze camera.
// The buffer that pRangeMap points to must be allocated accordingly before 
// passing it to the calculateRangeMap function.
void calculateRangeMap(const PartInfo& pointCloud, int minDepth, int maxDepth, uint16_t* pRangeMap, float invalidDataValue)
{
    const int width = (int)pointCloud.width;
    const int height = (int)pointCloud.height;
    CBlazeCamera::Coord3D *pPoint = reinterpret_cast<CBlazeCamera::Coord3D*>(pointCloud.pData);

    const double scale = 65535.0 / (maxDepth - minDepth);

    for (int row = 0; row < height; ++row)
    {
        for (int col = 0; col < width; ++col, ++pPoint, ++pRangeMap)
        {
            if (isValid(pPoint, invalidDataValue))
            {
                // calculate the radial distance
                double distance = sqrt(pPoint->x * pPoint->x + pPoint->y * pPoint->y + pPoint->z * pPoint->z);
                // clip to [minDepth..MaxDept]
                if (distance < minDepth)
                    distance = minDepth;
                else if (distance > maxDepth)
                    distance = maxDepth;
                *pRangeMap = (uint16_t) ( ( distance - minDepth ) * scale );
            }
            else
            {
                // No depth information available for this pixel. Zero it.
                *pRangeMap = 0;
            }
        }
    }
}

// Calculates a color depth map from point cloud data sent from a Basler blaze camera.
// The buffer that pRangeMap points to must be allocated accordingaly before
// passing it to the calculateRangeMap function.
#pragma pack(push, 1)
struct BGR 
{
    uint8_t b;
    uint8_t g;
    uint8_t r;
};
#pragma pack(pop)


void calculateRangeMapColor(const PartInfo& pointCloud, int minDepth, int maxDepth, BGR* pRangeMap, float invalidDataValue)
{
    const int width = (int) pointCloud.width;
    const int height = (int) pointCloud.height;
    CBlazeCamera::Coord3D *pPoint = reinterpret_cast<CBlazeCamera::Coord3D*>(pointCloud.pData);

    const double scale = 65535.0 / (maxDepth - minDepth);

    for (int row = 0; row < height; ++row)
    {
        for (int col = 0; col < width; ++col, ++pPoint, ++pRangeMap)
        {
            if (isValid(pPoint, invalidDataValue))
            {
                // calculate the radial distance
                double distance = sqrt(pPoint->x * pPoint->x + pPoint->y * pPoint->y + pPoint->z * pPoint->z);

                // clip to [minDepth..MaxDept]
                if (distance < minDepth)
                    distance = minDepth;
                else if (distance > maxDepth)
                    distance = maxDepth;
                
                // Calculate the color
                BGR bgr;
                const uint16_t g = (uint16_t)((distance - minDepth) * scale);
                const uint16_t val = g >> 6 & 0xff;
                const uint16_t sel = g >> 14;
                uint32_t res = val << 8 | 0xff;
                if (sel & 0x01)
                {
                    res = (~res) >> 8 & 0xffff;
                }
                if (sel & 0x02)
                {
                    res = res << 8;
                }
                bgr.r = res & 0xff;
                res = res >> 8;
                bgr.g = res & 0xff;
                res = res >> 8;
                bgr.b = res & 0xff;

                *pRangeMap = bgr;
            }
            else
            {
                // No depth information available for this pixel. Set it to black.
                BGR bgr;
                bgr.r = bgr.g = bgr.b = 0;
                *pRangeMap = bgr;
            }
        }
    }
}

class Sample
{
public:
    int run();

protected:
    bool onImageGrabbed(GrabResult grabResult, BufferParts);

private:
    CBlazeCamera    m_Camera;   // the blaze camera
    int             m_minDepth; // minimum distance value [mm]
    int             m_maxDepth; // maximum distance value [mm]
    const float     m_invalidDataValue = 0.0f; //  value that identifies an invalid pixel
};

int Sample::run()
{
    try
    {
        // Open the first camera found, i.e., establish a connection to the camera device.
        m_Camera.OpenFirstCamera();

        // Enable 3D (point cloud) data, intensity data, and confidence data.

        // Enable range data.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Range");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);

        // Range information can be sent either as a 16-bit gray value image or as 3D coordinates (point cloud).
        // For this sample, we want to acquire 3D coordinates.
        // Note: To change the format of an image component, the Component Selector must first be set to the component
        // you want to configure (see above).
        // To use 16-bit integer depth information, choose "Coord3D_C16" instead of "Coord3D_ABC32f".
        m_Camera.SetCameraParameterValue("PixelFormat", "Coord3D_ABC32f");
        // Set the value that identifies an invalid pixel. This value enables invalid pixels
        // to be identified and treated separately during image processing.
        m_Camera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateC");
        m_Camera.SetCameraParameterValue("Scan3dInvalidDataValue", (double) m_invalidDataValue);

        m_Camera.SetCameraParameterValue("ComponentSelector", "Intensity");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Mono16");

        m_Camera.SetCameraParameterValue("ComponentSelector", "Confidence");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Confidence16");

        // Query the minimum and maximum distance parameters.
        m_minDepth = (int) m_Camera.GetParameter<GenApi::CIntegerPtr>("DepthMin")->GetValue();
        m_maxDepth = (int)m_Camera.GetParameter<GenApi::CIntegerPtr>("DepthMax")->GetValue();

        // Prepare windows for displaying intensity image and depth map.
        cv::namedWindow("Intensity", cv::WINDOW_AUTOSIZE);
        cv::namedWindow("Confidence", cv::WINDOW_AUTOSIZE);
        cv::namedWindow("Range Map", cv::WINDOW_AUTOSIZE);
        cv::namedWindow("Range Map Color", cv::WINDOW_AUTOSIZE);

        std::cout << "To exit, press 'q' in one of the image windows" << std::endl;

        // Acquire and process images until the onImageGrabbed callback indicates to stop acquisition. 
        m_Camera.GrabContinuous(2, 1000, this, &Sample::onImageGrabbed);

        // Clean-up
        m_Camera.Close();
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

// Called for each frame grabbed.
bool Sample::onImageGrabbed(GrabResult grabResult, BufferParts parts)
{
    int key = 0;
    if (grabResult.status == GrabResult::Timeout)
    {
        std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
        return false; // Indicate to stop acquisition
    }
    if (grabResult.status != GrabResult::Ok)
    {
        std::cerr << "Image was not grabbed." << std::endl;
    }
    else
    {
        try
        {
            // -------  Create OpenCV images from grabbed buffer

            // First the intensity image...
            const int width = (int)parts[0].width;
            const int height = (int)parts[0].height;
            const int count = width * height;
            cv::Mat intensity = cv::Mat(height, width, CV_16UC1, parts[1].pData);
            // Scale the intensity image since it often looks quite dark
            double  max;
            cv::minMaxLoc(intensity, NULL, &max);
            intensity /= (max / std::numeric_limits<uint16_t>::max());

            // ... then the range images ...
            uint16_t* pRangeData = new uint16_t[width * height];
            calculateRangeMap(parts[0], m_minDepth, m_maxDepth, pRangeData, m_invalidDataValue);
            cv::Mat rangeMap = cv::Mat(height, width, CV_16UC1, pRangeData);
            
            BGR* pRangeDataColor = new BGR[width * height];
            calculateRangeMapColor(parts[0], m_minDepth, m_maxDepth, pRangeDataColor, m_invalidDataValue);
            cv::Mat3b rangeMapColor = cv::Mat(height, width, CV_8UC3, pRangeDataColor);

            // ... and third, the confidence map.
            cv::Mat confidence = cv::Mat(height, width, CV_16UC1, parts[2].pData);

            // -------   Display the images
            cv::imshow("Intensity", intensity);
            cv::imshow("Confidence", confidence);
            cv::imshow("Range Map", rangeMap);
            cv::imshow("Range Map Color", rangeMapColor);

            // ------ Since rangeMap and rangeMapColor didn't take ownership of the memory, 
            //        we must free the memory to prevent memory leaks.
            delete[] pRangeData;
            delete[] pRangeDataColor;
        }
        catch (const cv::Exception&  e)
        {
            std::cerr << e.what() << std::endl;
        }
    }
    key = cv::waitKey(1);
    return 'q' != (char)key;
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;
    try
    {
        CBlazeCamera::InitProducer();
        exitCode = Sample().run();
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    return exitCode;
}
