#include "FeatureDumper.h"
#include <string>
#include <ostream>

#include <boost/algorithm/string.hpp>

#include <ConsumerImplHelper/BlazeCamera.h>


using namespace GenICam;
using namespace GenApi;

void FeatureDumper::DumpTree(std::ostream &out, GenApi::INodeMap *pNodeMap)
{
    CNodeMapPtr ptrNodeMap(pNodeMap);
    assert(ptrNodeMap.IsValid());

    out << "# FeatureDumper\n";
    out << "# =============\n";

    CValuePtr ptrFeature;

    ptrFeature = ptrNodeMap->GetNode("DeviceVendorName");

    if (ptrFeature.IsValid() && IsReadable(ptrFeature)) {
        out << "# " << ptrFeature->GetNode()->GetName() << ":  " << ptrFeature->ToString() << "\n";
    }

    ptrFeature = ptrNodeMap->GetNode("DeviceModelName");

    if (ptrFeature.IsValid() && IsReadable(ptrFeature)) {
        out << "# " << ptrFeature->GetNode()->GetName() << ":  " << ptrFeature->ToString() << "\n";
    }

    ptrFeature = ptrNodeMap->GetNode("DeviceID");

    if (ptrFeature.IsValid() && IsReadable(ptrFeature)) {
        out << "# " << ptrFeature->GetNode()->GetName() << ":  " << ptrFeature->ToString() << "\n";
    }

    ptrFeature = ptrNodeMap->GetNode("DeviceVersion");

    if (ptrFeature.IsValid() && IsReadable(ptrFeature)) {
        out << "# " << ptrFeature->GetNode()->GetName() << ":  " << ptrFeature->ToString() << "\n";
    }

    ptrFeature = ptrNodeMap->GetNode("DeviceFirmwareVersion");

    if (ptrFeature.IsValid() && IsReadable(ptrFeature)) {
        out << "# " << ptrFeature->GetNode()->GetName() << ":  " << ptrFeature->ToString() << "\n";
    }

    out << "\n";

    CCategoryPtr ptrRoot;
    ptrRoot = ptrNodeMap->GetNode("Root");

    if (!ptrRoot.IsValid()) {
        throw RUNTIME_EXCEPTION("Root node not found");
    }
    DumpFeatures(ptrRoot, out, pNodeMap, 0);
}

void FeatureDumper::DumpFeatures(GenApi::ICategory *pCategory, std::ostream &out, GenApi::INodeMap *pNodeMap, int indent)
{

    CCategoryPtr ptrCategory(pCategory);
    assert(ptrCategory.IsValid());

    if (IsImplemented(ptrCategory)) {

        out << getIndent(indent) << ptrCategory->GetNode()->GetName() << ":" <<std::endl;
        out << getIndent(indent+INDENT) << "Name: " << ptrCategory->GetNode()->GetName() << std::endl;
        out << getIndent(indent+INDENT) << "DisplayName: " << ptrCategory->GetNode()->GetDisplayName() << std::endl;
        out << getIndent(indent+INDENT) << "Type: " << GetInterfaceName(ptrCategory) << std::endl;
        out << getIndent(indent+INDENT) << "AccessMode: " << getAccessString(ptrCategory->GetNode()->GetAccessMode()) << std::endl;
        out << getIndent(indent+INDENT) << "Visibility: " << getVisibilityString(ptrCategory->GetNode()->GetVisibility()) << std::endl;
        out << getIndent(indent+INDENT) << "ToolTip: " << ptrCategory->GetNode()->GetToolTip() << std::endl;
        out << getIndent(indent+INDENT) << "Description: " << ptrCategory->GetNode()->GetDescription() << std::endl;

        FeatureList_t FeatureList;
        ptrCategory->GetFeatures(FeatureList);

        for (FeatureList_t::const_iterator itFeature = FeatureList.begin(); itFeature != FeatureList.end(); ++itFeature) {
            CCategoryPtr ptrSubCategory(*itFeature);

            try {
                if (ptrSubCategory.IsValid()) {
                    DumpFeatures(ptrSubCategory, out, pNodeMap, indent+INDENT);
                } else {
                    DumpFeature(*itFeature, out, indent+INDENT);
                }
            } catch (...) {
                out << getIndent(indent) << "ERROR: Skipped Node/Feature" << std::endl;
            }

        }
    }
}


void FeatureDumper::DumpFeature(GenApi::IValue *pFeature, std::ostream &out, int indent)
{
    CValuePtr ptrFeature(pFeature);
    assert(ptrFeature.IsValid());

    // get the selector set
    CSelectorSet SelectorSet(ptrFeature);
    SelectorSet.SetFirst();


    // walk the selector set
    do {
        if(SelectorSet.ToString() != "[]") {
            // Transform   [DeviceTemperatureSelector=Illumination] to DeviceTemperatureSelector_Illumination
            std::string str = SelectorSet.ToString().c_str();
            boost::replace_all(str, "[", "");
            boost::replace_all(str, "]", "");
            boost::replace_all(str, "=", "_");

            out << getIndent(indent) << ptrFeature->GetNode()->GetName() << "_" << str << ":" << std::endl;

        } else {
            out << getIndent(indent) << ptrFeature->GetNode()->GetName() << ":" << std::endl;
        }


        indent+=INDENT;
        try {
            out << getIndent(indent) << "Name: " << ptrFeature->GetNode()->GetName() << std::endl;
            out << getIndent(indent) << "DisplayName: " << ptrFeature->GetNode()->GetDisplayName() << std::endl;
            out << getIndent(indent) << "Type: " << GetInterfaceName(ptrFeature) << std::endl;
            out << getIndent(indent) << "AccessMode: " << getAccessString(ptrFeature->GetNode()->GetAccessMode()) << std::endl;
            out << getIndent(indent) << "Visibility: " << getVisibilityString(ptrFeature->GetNode()->GetVisibility()) << std::endl;

            if(IsReadable(ptrFeature)) {
                out << getIndent(indent) << "Value: " << ptrFeature->ToString() << std::endl;

                switch (ptrFeature->GetNode()->GetPrincipalInterfaceType()) {
                case intfIInteger: {
                    CIntegerPtr ptrInteger(ptrFeature);
                    out << getIndent(indent) << "Min: " << ptrInteger->GetMin() << std::endl;
                    out << getIndent(indent) << "Max: " << ptrInteger->GetMax() << std::endl;
                    out << getIndent(indent) << "Inc: " << ptrInteger->GetInc() << std::endl;

                    if (ptrInteger->GetUnit() != "") {
                        out << getIndent(indent) << "Unit: " << ptrInteger->GetUnit() << std::endl;
                    }
                }
                break;

                case intfIFloat: {
                    CFloatPtr ptrFloat(ptrFeature);
                    out << getIndent(indent) << "Min: " << ptrFloat->GetMin() << std::endl;
                    out << getIndent(indent) << "Max: " << ptrFloat->GetMax() << std::endl;

                    if (ptrFloat->HasInc()) {
                        out << getIndent(indent) << "Inc: " << ptrFloat->GetInc() << std::endl;
                    }

                    if (ptrFloat->GetUnit() != "") {
                        out << getIndent(indent) << "Unit: " << ptrFloat->GetUnit() << std::endl;
                    }
                }
                break;

                case intfIEnumeration: {
                    CEnumerationPtr ptrEnumeration(ptrFeature);
                    GenApi::StringList_t Symbolics;
                    ptrEnumeration->GetSymbolics(Symbolics);

                    bool First = true;
                    out << getIndent(indent) << "Symbolics: {";
                    GenApi::StringList_t::iterator itSymbolic;

                    for (itSymbolic = Symbolics.begin(); itSymbolic != Symbolics.end(); itSymbolic++) {
                        if (First) {
                            First = false;
                        } else {
                            out << ", ";
                        }

                        out << (*itSymbolic).c_str();
                    }

                    out << "}" << std::endl;
                }
                break;

                // the rest falls through
                case intfIValue:
                case intfIString:
                case intfICategory:
                case intfIRegister:
                case intfIBoolean:
                case intfIBase:
                case intfICommand:
                case intfIEnumEntry:
                case intfIPort:
                    break;
                }

            }

            out << getIndent(indent) << "ToolTip: " << ptrFeature->GetNode()->GetToolTip() << std::endl;
            out << getIndent(indent) << "Description: \"" << ptrFeature->GetNode()->GetDescription() << "\""<< std::endl;
        } catch (GenICam::GenericException& e) {
            out << getIndent(indent) << "ERROR: Failed to dump feature: " << e.GetDescription() << std::endl;
        } catch (...) {
            out << getIndent(indent) << "ERROR: Failed to dump feature." << std::endl;
        }
        indent-=INDENT;
    } while (SelectorSet.SetNext());

    try {
        SelectorSet.Restore();
    } catch (std::exception &e) {
        std::cout << "ERROR : " << e.what() << "\n";
        std::cerr << "ERROR : " << e.what() << "\n";
    }

}

std::string FeatureDumper::getIndent(int indent)
{
    std::string str = "";
    for(int i=0; i<indent; i++) {
        str += " ";
    }
    return str;
}

std::string FeatureDumper::getAccessString(GenApi::EAccessMode val)
{
    switch(val) {
    case NI:
        return "NI";
        break;
    case NA:
        return "NA";
        break;
    case WO:
        return "WO";
        break;
    case RO:
        return "RO";
        break;
    case RW:
        return "RW";
        break;
    case _UndefinedAccesMode:
        return "_UndefinedAccesMode";
        break;
    case _CycleDetectAccesMode:
        return "_CycleDetectAccesMode";
        break;
    default:
        return "";
    }
}

std::string FeatureDumper::getVisibilityString(GenApi::EVisibility val)
{
    switch(val) {
    case Beginner:
        return "Beginner";
        break;
    case Expert:
        return "Expert";
        break;
    case Guru:
        return "Guru";
        break;
    case Invisible:
        return "Invisible";
        break;
    default:
        return "_UndefinedVisibility";
    }
}
