/*
This sample illustrates how to grab depth maps from a blaze camera.

Refer to the ConvertPointCloud2DepthMap sample for how to apply a color map
to depth maps.

OpenCV is used only for displaying the depth maps.

*/

#include <iostream>
#include <iomanip> 
#include <atomic>
#include <opencv2/opencv.hpp>
#include <ConsumerImplHelper/BlazeCamera.h>

using namespace GenTLConsumerImplHelper;

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    const int missingDepth = 0;  // gray value reserved for indicating missing depth data. 

    try
    {
        CBlazeCamera::InitProducer();

        CBlazeCamera camera;
        camera.OpenFirstCamera();
        std::cout << "Connected to camera " << camera.GetCameraInfo().strDisplayName << std::endl;

        // Enable depth maps by enabling the Range component and setting the appropriate pixel format.
        camera.SetCameraParameterValue("ComponentSelector", "Range");
        camera.SetCameraParameterValue("ComponentEnable", true);
        camera.SetCameraParameterValue("PixelFormat", "Coord3D_C16");

        // Query the conversion factor required to convert gray values to distances:
        // Choose the z axis first...
        camera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateC");
        // ... then retrieve the conversion factor.
        const auto gray2mm = camera.GetCameraParameterValue<double>("Scan3dCoordinateScale");

        // Configure the gray value used for indicating missing depth data.
        // Note: Before setting the value, Scan3dCoordinateSelector must be set to the axis the 
        // value is to be configured for. In this case the z axis, i.e., Scan3dCoordianteSelector, must be set 
        // to "CoordinateC", which already has been done a few lines above.
        camera.SetCameraParameterValue("Scan3dInvalidDataValue", (double) missingDepth); 

        // Retrieve calibration data from the camera.
        const auto cx = camera.GetCameraParameterValue<double>("Scan3dPrincipalPointU");
        const auto cy = camera.GetCameraParameterValue<double>("Scan3dPrincipalPointV");
        const auto f = camera.GetCameraParameterValue<double>("Scan3dFocalLength");

        // Prepare window for displaying the depth map.
        cv::namedWindow("Depth Map", cv::WINDOW_AUTOSIZE);

        std::cout << "To exit, press 'q' or 'ESC' in the image windows" << std::endl;

        // Since called by a different thread, the following lambda function must capture
        // variables by value only!
        camera.GrabContinuous(5, 1000, [=](GrabResult grabResult, BufferParts parts) -> bool {
            if (grabResult.status == GrabResult::Timeout)
            {
                std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
                return false; // Indicate to stop acquisition
            }
            if (grabResult.status != GrabResult::Ok)
            {
                std::cerr << "Image was not grabbed successfully." << std::endl;
            }
            else
            {
                // Retrieve the center pixel's gray value.
                const auto width = parts[0].width;
                const auto height = parts[0].height;
                const int u = (int)(0.5 * width);
                const int v = (int)(0.5 * height);
                uint16_t g = ((uint16_t*) parts[0].pData)[u + v * width];
                if (g != missingDepth)
                {
                    // Calculate 3D coordinate from gray value and pixel coordinate.
                    const double z = g * gray2mm;
                    const double x = (u - cx) * z / f;
                    const double y = (v - cy) * z / f;

                    std::cout.setf(std::ios_base::fixed);
                    std::cout.precision(1);
                    std::cout << "x=" << std::setw(6) << x << "mm y=" << std::setw(6) << y << "mm z=" << std::setw(6) << z << "mm \n";
                }
                else
                {
                    std::cout << "x=   n/a   y=   n/a   z=   n/a\n";
                }
                // Display the depth map.
                cv::Mat depth = cv::Mat((int)height, (int)width, CV_16UC1, parts[0].pData);
                cv::imshow("Depth Map", depth);
            }
            int key = cv::waitKey(1);
            return 'q' != (char)key && 27 != key;  // ESC or q to quit
            });

        // GrabContinuous() blocks execution until the grab callback returns false.

        // Clean-up
        camera.Close();
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}
