/*
This sample illustrates how to operate several blaze cameras simultaneously
without their light sources interfering with each other.

The cameras are operated in free run mode. A typical application would be,
e.g., in automated guided vehicles (AGVs), which operate independently from each
and should not interfere with each other.

Refer to the MultiCam/SynchronousFreeRun sample to learn how the clocks of
multiple cameras can be synchronized precisely via PTP.

ATTENTION: The MultiCameraChannel feature requires cameras with hardware 
revision 8 or above.

OpenCV is only used for displaying the depth maps.

*/

#include <iostream>
#include <iomanip>
#include <thread>
#include <ConsumerImplHelper/BlazeCamera.h>
#if defined WITH_OPENCV
#  include <opencv2/opencv.hpp>
#endif

using namespace GenTLConsumerImplHelper;

class Sample
{
public:
    int run(CameraInfo cameraInfo, int cameraNumber);

    bool onImageGrabbed(GrabResult grabResult, BufferParts);

private:
    CBlazeCamera m_Camera;
    int          m_nBuffersGrabbed;
};

int Sample::run(CameraInfo cameraInfo, int cameraNumber)
{
    m_nBuffersGrabbed = 0;

    try {
        // Open the camera, i.e., establish a connection to the camera device.
        m_Camera.Open(cameraInfo);

        std::cout << "Connected to camera " << m_Camera.GetCameraInfo().strDisplayName << std::endl;

        // Set a unique channel for each camera that is different from the others
        int channel = cameraNumber % (m_Camera.GetParameter<GenApi::CIntegerPtr>("MultiCameraChannel")->GetMax() + 1);
        m_Camera.SetCameraParameterValue("MultiCameraChannel", channel);

        // Enable depth maps by enabling the Range component and setting the appropriate pixel format.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Range");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Coord3D_C16");

        // Acquire images until the onImageGrabbed callback indicates to stop acquisition.
        // 5 buffers are used (round-robin).
        m_Camera.GrabContinuous(5, 1000, this, &Sample::onImageGrabbed);

        // Clean-up
        m_Camera.Close();
    } catch (const GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used
        // to check if the device is still connected.
        if (m_Camera.IsOpen() && !m_Camera.IsConnected()) {
            std::cerr << "Camera has been removed." << std::endl;
        }

        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

bool Sample::onImageGrabbed(GrabResult grabResult, BufferParts parts)
{
    if (grabResult.status == GrabResult::Timeout) {
        std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
        // The timeout might be caused by a removal of the camera. Check if the camera
        // is still connected.
        if (!m_Camera.IsConnected()) {
            std::cerr << "Camera has been removed." << std::endl;
        }

        return false; // Indicates that acquisition will be stopped.
    }
    m_nBuffersGrabbed++;
    if (grabResult.status != GrabResult::Ok) {
        std::cerr << "Image " << m_nBuffersGrabbed << " was not grabbed." << std::endl;
    } else {
        // Retrieve the values for the center pixel.
        const int width = (int)parts[0].width;
        const int height = (int)parts[0].height;
  
        std::cout << m_Camera.GetCameraInfo().strDisplayName << " - image nr. " << m_nBuffersGrabbed << std::endl;

#if defined WITH_OPENCV
        cv::Mat depth = cv::Mat(height, width, CV_16UC1, parts[0].pData);

        // -------   Display the images.
        cv::imshow(m_Camera.GetCameraInfo().strDisplayName, depth);
        cv::waitKey(1);
#endif

    }
    return m_nBuffersGrabbed < 30; // Indicates that acquisition will be stopped when 30 buffers have been grabbed.
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try {
        CBlazeCamera::InitProducer();

        CameraList cameraList = CBlazeCamera::EnumerateCameras();
        if (cameraList.size() == 0) {
            std::cout << "No camera found!" << std::endl;
            exitCode = EXIT_FAILURE;
        }
        int camCnt = 0;
        std::vector<std::thread> cameraThreads;
        for (auto& cameraInfo : cameraList) {
            cameraThreads.emplace_back([&] { Sample().run(cameraInfo, camCnt++); });
        }
        for (auto& thread : cameraThreads) {
            thread.join();
        }
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources.
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized()) {
        CBlazeCamera::TerminateProducer();    // Won't throw any exceptions.
    }

    return exitCode;
}
