/*
This sample illustrates how to operate several blaze cameras simultaneously
without their light sources interfering with each other.

The cameras are triggered using software triggers. The trigger is used to
operate the cameras synchronously and to obtain images from approximately the
same point in time. A typical use case would be, e.g., to merge the
images of multiple cameras into a common point cloud to obtain a complete 
3D model of an object or scene. If synchronous operation isn't required, there 
is no need for using software triggers. 

Refer to the MultiCam/SynchronousFreeRun sample to learn how the clocks of
multiple cameras can be synchronized precisely via PTP.

ATTENTION: The MultiCameraChannel feature requires cameras with hardware 
revision 8 or above. If your camera doesn't support that feature, an error 
message will be printed out.

OpenCV is only used for displaying the depth maps.

*/

#include <iostream>
#include <iomanip>
#include <atomic>
#include <ConsumerImplHelper/BlazeCamera.h>
#if defined WITH_OPENCV
#  include <opencv2/opencv.hpp>
#endif

using namespace GenTLConsumerImplHelper;

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;
    const int nBuffers = 2;   // Number of buffers used per camera.
    unsigned int idxNextBuffer = 0;
    int key = 0;

    std::unique_ptr< std::vector<CBlazeCamera>> cameras;

    try {
        CBlazeCamera::InitProducer();

        CameraList cameraList = CBlazeCamera::EnumerateCameras();
        const size_t numCameras = cameraList.size();
        if (numCameras == 0) {
            std::cout << "No camera found!" << std::endl;
            exitCode = EXIT_FAILURE;
        }

        cameras = std::make_unique<std::vector<CBlazeCamera>>(numCameras);
        int camCnt = 0;
        for (auto cameraInfo : cameraList) {
            auto& camera = cameras->at(camCnt);
            camera.Open(cameraInfo);
            std::cout << "Connected to camera " << camera.GetCameraInfo().strDisplayName << std::endl;

            // Set a unique channel for each camera that is different from the others
            int channel = camCnt % (camera.GetParameter<GenApi::CIntegerPtr>("MultiCameraChannel")->GetMax() + 1);
            camera.SetCameraParameterValue("MultiCameraChannel", channel);

            // Enable depth maps by enabling the Range component and setting the appropriate pixel format.
            camera.SetCameraParameterValue("ComponentSelector", "Range");
            camera.SetCameraParameterValue("ComponentEnable", true);
            camera.SetCameraParameterValue("PixelFormat", "Coord3D_C16");

            // Configure the camera for software triggering.
            camera.SetCameraParameterValue("TriggerMode", "On");
            camera.SetCameraParameterValue("TriggerSource", "Software");

            // Allocate memory buffers for grabbing.
            camera.PrepareAcquisition(nBuffers);

            // For each camera, enqueue one buffer that will receive the image data of the next image
            // the corresponding camera sends.
            camera.QueueBuffer(idxNextBuffer);

            // Start the acquisition on the host. Now, the enqueued buffers will be filled with data sent by the cameras.
            camera.StartAcquisition();

            // The cameras won't send any data until they have been started.
            camera.IssueAcquisitionStartCommand();

            // Issue software trigger.
            camera.ExecuteCameraCommand("TriggerSoftware");

#if defined WITH_OPENCV
            // Prepare window for displaying the depth map.
            cv::namedWindow(camera.GetCameraInfo().strDisplayName, cv::WINDOW_AUTOSIZE);
#endif

            camCnt++;
        }

#if defined WITH_OPENCV
        std::cout << "To exit, press 'q' or 'ESC' in the image windows." << std::endl;
#endif

        // Grab loop. Acquire and process data from all cameras.
        // While the data is being processed for each camera, a new buffer will be filled in
        // the background.
        int nBuffersGrabbed = 0;
        while (exitCode == EXIT_SUCCESS) {
            nBuffersGrabbed++;
            idxNextBuffer++;
            idxNextBuffer %= nBuffers;

            // Process n images from each camera.
            for (auto& camera : *cameras) {
                GrabResult grabResult;
                camera.GetGrabResult(grabResult, 1000);

                // Enqueue the next buffers.
                // The new buffers will be filled asynchronously in the background.
                // Each buffer will be filled with the next image received from the corresponding camera.

                camera.QueueBuffer(idxNextBuffer);
                camera.ExecuteCameraCommand("TriggerSoftware");

                // While the next buffers are acquired, process the current ones.
                if (grabResult.status != GrabResult::Ok) {
                    std::cerr << "Failed to grab image from camera " << camera.GetCameraInfo().strDisplayName << std::endl;
                    if (grabResult.status == GrabResult::Timeout) {
                        std::cerr << "Timeout occurred." << std::endl;
                    }
                    exitCode = EXIT_FAILURE;
                }

                if (exitCode == EXIT_SUCCESS) {

                    // Successfully acquired the image.
                    BufferParts parts;
                    camera.GetBufferParts(grabResult, parts);

                    const int width = (int)parts[0].width;
                    const int height = (int)parts[0].height;
                    const int count = width * height;

                    std::cout << camera.GetCameraInfo().strDisplayName << " - image nr. " << nBuffersGrabbed << std::endl;

#if defined WITH_OPENCV
                    cv::Mat depth = cv::Mat(height, width, CV_16UC1, parts[0].pData);

                    // -------   Display the images.
                    cv::imshow(camera.GetCameraInfo().strDisplayName, depth);
#endif
                }
            }

#if defined WITH_OPENCV
            key = cv::waitKey(1);
            if ('q' == (char)key) {
                break;
            }
#else
            if (nBuffersGrabbed >= 30) {
                break;
            }
#endif

        }
    } catch (GenICam::GenericException& e) {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // clean-up
    if (cameras) {
        for (auto& camera : *cameras) {
            if (camera.IsOpen()) {
                try {
                    if (camera.IsAcquisitionPrepared()) {
                        // Stop acquisition and free all grab-related resources.
                        camera.FinishAcquisition();
                    }

                    // Disable software trigger.
                    camera.SetCameraParameterValue("TriggerMode", "Off");

                    camera.Close();
                }
                catch (...)
                {
                }
            }
        }
        cameras.reset();  // calls destructors of CBlazeCamera objects.
    }

    // Release the GenTL producer and all of its resources.
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized()) {
        CBlazeCamera::TerminateProducer();    // Won't throw any exceptions.
    }

    return exitCode;
}
