#if !defined(NOMINMAX)
#  define NOMINMAX
#endif

#include <deque>
#include <thread>
#include <functional>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include <numeric>
#include <cmath>

#include "PtpHelper.h"

using namespace std::chrono_literals;
using namespace GenTLConsumerImplHelper;

namespace blazePtpHelper
{
    std::ostream* debugStream = &std::cout;
    std::ostream* errorStream = &std::cerr;

    class NullStream : public std::ostream 
    {
    public:
        NullStream() : std::ostream(&m_buffer) {}
    private:
        class NullBuf : public std::streambuf {
        public:
            virtual std::streamsize xsputn(const char*, std::streamsize n)
            {
                return n;
            }
            virtual int overflow(int) {
                return 1;
            }
        } m_buffer;
    } nullStream;

    void enableLogging(bool enable)
    {
        if (enable)
        {
            debugStream = &std::clog;
            errorStream = &std::cerr;
        }
        else
        {
            debugStream = &nullStream;
            errorStream = &nullStream;
        }
    }

    void enablePtp(BlazeCamera & camera)
    {
        camera.SetCameraParameterValue("PtpEnable", true);
     }

    bool openCamerasAndEnablePtp(BlazeCameraArray& cameras)
    {
        cameras.clear();
        // Search for cameras.
        auto cameraInfoList = CBlazeCamera::EnumerateCameras();
        if (cameraInfoList.empty())
        {
            error() << "No cameras found." << std::endl;
            return false;
        }
        cameras.reserve(cameraInfoList.size());

        // Establish connection to cameras and enable PTP.
        for (const auto& deviceInfo : cameraInfoList)
        {
            auto ptrCamera = std::make_unique<CBlazeCamera>();
            try
            {
                ptrCamera->Open(deviceInfo);
            }
            catch (const GenICam::GenericException& e)
            {
                error() << deviceInfo <<
                    " Failed to open camera. : " << e.GetDescription() << std::endl;
                error() << deviceInfo << " skipped." << std::endl;
            }
            if (ptrCamera->IsOpen())
            {
                try
                {
                    enablePtp(*ptrCamera); // enable ptp
                    cameras.push_back(std::move(ptrCamera)); // camera array now has ownership
                }
                catch (const GenICam::GenericException& e)
                {
                    error() << deviceInfo
                        << " Ignoring camera. Reason:\n" << e.GetDescription() << std::endl;
                }
            }
        }
        return true;
    }

    bool waitForMasterSlaveNegotation(const BlazeCameraArray& cameras, std::chrono::seconds timeout)
    {
        using clock = std::chrono::steady_clock;
        const auto t0 = clock::now();
        debug() << "Waiting until all cameras negotiated their PTP role..." << std::endl;
        size_t nMasters;
        size_t nSlaves;
        bool done = false;
        bool timedOut = false;
        do
        {
            nMasters = 0;
            nSlaves = 0;
            auto t1 = clock::now();
            for (const auto& camera : cameras)
            {
                latchPtpStatus(*camera);
                const auto currentStatus = getPtpStatus(*camera);
                debug() << *camera << " Ptp Status = " << currentStatus << std::endl;
                if (currentStatus.compare("Master") == 0) nMasters++;
                if (currentStatus.compare("Slave") == 0) nSlaves++;
            }
            debug() << "-------" << std::endl;
            done = (nMasters + nSlaves) == cameras.size() && nMasters <= 1;
            timedOut = clock::now() - t0 > timeout;
            if (!timedOut && !done)
            {
                std::this_thread::sleep_until(t1 + std::chrono::microseconds{ 500 });
            }
        } while (! ( done || timedOut) );

        return done;
    }

    bool waitForClocksSynchronization   (const BlazeCameraArray& cameras,
        std::chrono::nanoseconds threshold, 
        std::chrono::milliseconds window, 
        std::chrono::seconds timeout)
    {
        using clock = std::chrono::steady_clock;
        debug() << "Waiting until all clocks are in sync..." << std::endl;

        std::vector<std::chrono::nanoseconds> offsetsFromMaster;
        offsetsFromMaster.resize(cameras.size());
        const auto t0 = clock::now(); // for controlling the timeout
        auto windowStart = t0; // for measuring the window
        bool windowEndReached = false;
        bool timedOut = false;
        do
        {
            offsetsFromMaster.clear();
            for (const auto& camera : cameras)
            {
                latchPtpStatus(*camera);
                offsetsFromMaster.push_back(getOffsetFromMaster(*camera));
            }

            const auto maxOffsetFromMaster = *std::max_element(offsetsFromMaster.begin(), offsetsFromMaster.end(), 
                    [](std::chrono::nanoseconds a, std::chrono::nanoseconds b) { return abs(a.count()) < abs(b.count()); });
            debug() << "Max offset from master: " << maxOffsetFromMaster.count() << "ns\n";
            if ( std::chrono::nanoseconds(abs(maxOffsetFromMaster.count())) > threshold)
            {
                // reset time window
                windowStart = clock::now();
            }
            const auto t1 = clock::now();
            timedOut = t1 - t0 > timeout;
            windowEndReached = t1 - windowStart > window;
        } while (!windowEndReached && !timedOut);
        
        return windowEndReached;
    }

    void latchPtpStatus(BlazeCamera& camera)
    {
        try
        {
            // Update ptp related values.
            camera.ExecuteCameraCommand("PtpDataSetLatch");
        }
        catch (const GenICam::GenericException& ex)
        {
            throw RUNTIME_EXCEPTION("%s Failed to latch ptp data set: %s", formatCameraName(camera).c_str(), ex.GetDescription());
        }
    }
    
    std::string getPtpStatus(BlazeCamera& camera)
    {
        return camera.GetCameraParameterValue<std::string>("PtpStatus");
    }

    std::chrono::nanoseconds getOffsetFromMaster(BlazeCamera& camera)
    {
        return std::chrono::nanoseconds(camera.GetCameraParameterValue<int64_t>("PtpOffsetFromMaster"));
    }

    std::chrono::nanoseconds latchAndGetTimestamp(BlazeCamera& camera)
    {
        camera.ExecuteCameraCommand("TimestampLatch");
        return std::chrono::nanoseconds(camera.GetCameraParameterValue<int64_t>("TimestampLatchValue"));
    }

    std::chrono::microseconds getExposureTime(BlazeCamera& camera)
    {
        return std::chrono::microseconds((int64_t)camera.GetCameraParameterValue<double>("ExposureTime"));
    }

    std::chrono::microseconds getReadoutTime(BlazeCamera& camera)
    {
        return std::chrono::microseconds(camera.GetCameraParameterValue<int64_t>("ReadoutTime"));
    }

    std::chrono::microseconds getResetTime(BlazeCamera& camera)
    {
        return std::chrono::microseconds(camera.GetCameraParameterValue<int64_t>("ResetTime"));
    }

    std::chrono::microseconds getStartupTime(BlazeCamera& camera)
    {
        return std::chrono::microseconds(camera.GetCameraParameterValue <int64_t>("StartupTime"));
    }

    bool isFastModeEnabled(BlazeCamera& camera)
    {
        return camera.GetCameraParameterValue<bool>("FastMode");
    }

    std::chrono::microseconds getFrameDuration(BlazeCamera& camera)
    {
        const auto frameDuration = camera.GetCameraParameterValue<int64_t>("FrameDuration");
        return std::chrono::microseconds(frameDuration);
    }

    double calculateMaxFrameRate(const BlazeCameraArray& cameras, bool interleaved,
        std::chrono::microseconds safetyMargin)
    {
        std::chrono::microseconds totalFramePeriod;
        if (!interleaved)
        {
            const size_t n = cameras.size();
            totalFramePeriod = std::accumulate(
                cameras.begin(), cameras.end(), std::chrono::microseconds(n * safetyMargin), [safetyMargin](const auto& value, const auto& ptrCamera)
                {
                    const auto startupTime = getStartupTime(*ptrCamera);
                    const auto resetTime = getResetTime(*ptrCamera);
                    const auto readoutTime = getReadoutTime(*ptrCamera);
                    return value + getFrameDuration(*ptrCamera) - readoutTime - resetTime - startupTime;
                });
             totalFramePeriod += getStartupTime(*cameras.at(0)) + getResetTime(*cameras.at(0));
        }
        else 
        {
            totalFramePeriod = getFrameDuration(*cameras.at(0)) +safetyMargin;
        }

        const double frameRate = 1000000.0 / totalFramePeriod.count();

        // Determine the lowest maximum allowed frame rate.
        double lowestMaxFrameRate = 1e9;
        for (auto& camera : cameras)
        {
            lowestMaxFrameRate = std::min(lowestMaxFrameRate, camera->GetParameter<GenApi::CFloatPtr>("AcquisitionFrameRate", GenApi::RO)->GetMax());
        }

        return std::min(lowestMaxFrameRate, frameRate);
    }

    bool setupCamerasForSynchronousFreeRun(const BlazeCameraArray& cameras, 
        std::chrono::microseconds safetyMargin)
    {
        // When there are two cameras, the second camera can perform the exposure while
        // the first camera reads out the sensor data.
        // In the following we call this timing scheme "interleaved".
        // This read out phase between exposures is not long enough to allow more than
        // two cameras using the interleaved timing.
        // When there are more than 2 cameras, all cameras are operated in a round robin
        // schedule.
        // ** NOTE: For using the interleaved timing, the cameras' exposure time
        // and FastMode settings must the same for each camera! **
        const bool useInterleavedTiming = cameras.size() == 2;

        const auto frameRate = calculateMaxFrameRate(cameras, useInterleavedTiming, safetyMargin);

        // Iterate over all cameras. Setup each camera's time to start.
        // To prevent cameras from sending out light at the same time, the cameras start exposure 
        // in a chronological sequence. 
        for (auto& camera : cameras)
        {
            // Enable fixed acquisition frame rate
            camera->SetCameraParameterValue<bool>("AcquisitionFrameRateEnable", true);
            camera->SetCameraParameterValue<double>("AcquisitionFrameRate", frameRate);
        }

        // Get the current ptp time and determine the point of time the synchronous free run
        // of the first camera should start.
        const std::chrono::nanoseconds now = latchAndGetTimestamp(*cameras.at(0));
        const auto timeToConfigureASingleCamera = 400ms;
        std::chrono::nanoseconds startAt = now + timeToConfigureASingleCamera * cameras.size();
        auto triggerOffset = std::chrono::microseconds(0);
        for ( size_t i = 0; i < cameras.size(); ++i )
        {
            auto& camera = cameras.at(i);
            // Set the point in time when the cameras should start the synchronous free run.
            camera->SetCameraParameterValue<int64_t>("SyncFreeRunTimerStartTimeLow", startAt.count() & 0xffffffff);
            camera->SetCameraParameterValue<int64_t>("SyncFreeRunTimerStartTimeHigh", startAt.count() >> 32);
            debug() << *camera << " Start time = "
                << startAt << " Frame rate = "
                << camera->GetCameraParameterValue<double>("AcquisitionFrameRate") << " fps.";
            debug() << " Trigger offset = " << std::chrono::duration_cast<std::chrono::microseconds> (triggerOffset).count() << "us." << std::endl;

            camera->ExecuteCameraCommand("SyncFreeRunTimerUpdate");
            camera->SetCameraParameterValue("SyncFreeRunTimerEnable", true);
            if (i < cameras.size() - 1)
            {
                if (useInterleavedTiming)
                {
                    triggerOffset = safetyMargin + getExposureTime(*camera);
                }
                else
                {
                    auto& nextCam = cameras.at(i + 1);
                    triggerOffset = safetyMargin + getFrameDuration(*camera) 
                        - getReadoutTime(*camera) - getStartupTime(*nextCam) - getResetTime(*nextCam);
                }
                startAt += triggerOffset;
            }
        }

        return true;
    }

    std::string formatCameraName(const GenTLConsumerImplHelper::CameraInfo& info)
    {
        std::stringstream s;
        s << info;
        return s.str();
    }

    std::string formatCameraName(const BlazeCamera& camera)
    {
        return formatCameraName(camera.GetCameraInfo());
    }

    MatchingQueues::MatchingQueues(size_t nQueues, size_t maxQueueLen)
        : m_nQueues(nQueues)
        , m_maxQueueLen(maxQueueLen)
        , m_queues(nQueues)
    {

    }

    MatchingQueues::~MatchingQueues()
    {
        discardAll();
    }

    void MatchingQueues::setDiscardCallback(DiscardCallback cb)
    {
        m_discardCallback = cb;
    }

    void MatchingQueues::setCompletionCallback(CompletionCallback cb)
    {
        m_completionCallback = cb;
    }

    void MatchingQueues::pushBack(size_t queueIdx, QueueItemBase* item)
    {
        if (queueIdx > m_nQueues)
        {
            throw std::out_of_range("Queue index out of range.");
        }
        auto& queue = m_queues[queueIdx];
        auto ptr = QueueItemPtr(item);
        queue.push_back(ptr);
        while ( queue.size() > m_maxQueueLen )
        {
            discardItem(queueIdx, queue.front());
            queue.pop_front();
        }
        update();
    }

    bool MatchingQueues::hasCompleteSet() const
    {
        return m_state == State::complete;
    }

    ItemSet MatchingQueues::popSet()
    {
        ItemSet result;
        if (m_state == State::complete)
        {
            result.reserve(m_nQueues);
            for (auto& q : m_queues)
            {
                result.push_back(q.front());
                q.pop_front();
            }
            update();
        }
        return result;
    }

    bool MatchingQueues::isThereNonEmptyQueue() const
    {
        return std::any_of(m_queues.begin(), m_queues.end(), [](const auto& q) { return !q.empty(); });
    }

    bool MatchingQueues::isThereEmptyQueue() const
    {
        return std::any_of(m_queues.begin(), m_queues.end(), [](const auto& q) { return q.empty(); });
    }

    void MatchingQueues::discardItem(size_t idx, const QueueItemPtr& item)
    {
        if (m_discardCallback)
        {
            m_discardCallback(idx, item.get());
        }
    }

    void MatchingQueues::discardAll()
    {
        while (isThereNonEmptyQueue())
        {
            size_t idx = 0;
            for (auto& queue : m_queues)
            {
                if (!queue.empty())
                {
                    discardItem(idx, queue.front());
                    queue.pop_front();
                }
                idx++;
            }
        }
    }

    void MatchingQueues::discardItemsLessThan(key_t key)
    {
        int nDiscarded = 0;
        do
        {
            nDiscarded = 0;
            size_t idx = 0;
            for (auto& queue : m_queues)
            {
                if (!queue.empty() && queue.front()->getKey() < key)
                {
                    nDiscarded++;
                    discardItem(idx, queue.front());
                    queue.pop_front();
                }
                idx++;
            }
        } while (nDiscarded > 0);
    }

    void MatchingQueues::update()
    {
        if (!isThereEmptyQueue())
        {
            auto minmax = getMinMaxKey();
            if (minmax.first == minmax.second)
            {
                // all front elements have the same key -> complete
                if (m_state != State::complete)
                {
                    m_state = State::complete;
                    if (m_completionCallback) {
                        m_completionCallback(*this);
                    }
                }
            }
            else
            {
                m_state = State::incomplete;
                discardItemsLessThan(minmax.second);
                // after discarding elements, there might be now a complete set,
                // so re-run update()
                update();
            }
        }
        else
        {
            m_state = State::incomplete;
        }
    }

    std::pair<MatchingQueues::key_t, MatchingQueues::key_t> MatchingQueues::getMinMaxKey() const
    {
        auto min = std::numeric_limits<key_t>::max();
        auto max = std::numeric_limits<key_t>::min();
        for (const auto& q : m_queues)
        {
            if (!q.empty())
            {
                const auto key = q.front()->getKey();
                if (key < min) {
                    min = key;
                }
                if (key > max) {
                    max = key;
                }
            }
        }
        return std::pair<key_t, key_t>(min, max);
    }


    ///////////////////////////////////////////////////////////////////////////////////////////////////////////////


    class QueueItemGrabResult : public QueueItemBase
    {
    public:
        QueueItemGrabResult(const GrabResult& grabResult)
            : m_grabResult(grabResult)
        {
        }

    public:
        virtual MatchingQueues::key_t getKey() override
        {
            return m_grabResult.frameID;
        }

        GenTLConsumerImplHelper::GrabResult grabResult() const
        {
            return m_grabResult;
        }

    private:
        GrabResult m_grabResult;
    };

    GrabResultMatcher::GrabResultMatcher(size_t nCameras, size_t maxQueueLen)
    {
        m_queueMatcher = std::make_unique<MatchingQueues>(nCameras, maxQueueLen);
        m_queueMatcher->setCompletionCallback(std::bind(&GrabResultMatcher::onComplete, this, std::placeholders::_1));
        m_queueMatcher->setDiscardCallback(std::bind(&GrabResultMatcher::onDiscard, this, std::placeholders::_1, std::placeholders::_2));
    }

    GrabResultMatcher::~GrabResultMatcher()
    {
        if (m_completionCallback)
            m_completionCallback = CompletionCallback();
        if (m_discardCallback)
            m_discardCallback = DiscardCallback();
        discardAllImpl();
    }

    void GrabResultMatcher::addGrabResult(size_t cameraIdx, const GrabResult& grabResult)
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        m_queueMatcher->pushBack(cameraIdx, new QueueItemGrabResult(grabResult));
    }

    bool GrabResultMatcher::hasCompleteSet() const
    {
        bool result = false;
        std::unique_lock<std::mutex> lock(m_mutex);
        result = m_queueMatcher->hasCompleteSet();
        return result;
    }

    GrabResultArray GrabResultMatcher::getGrabResults()
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        GrabResultArray result;
        auto set = m_queueMatcher->popSet();
        if (!set.empty())
        {
            result.reserve(set.size());
            for (auto& item : set)
            {
                result.push_back(dynamic_cast<QueueItemGrabResult&>(*item).grabResult());
            }
        }
        return result;
    }

    void GrabResultMatcher::setCallbacks(CompletionCallback cbCompletion, DiscardCallback cbDiscard, const void* pContext)
    {
        m_pUserContext = pContext;
        m_completionCallback = cbCompletion;
        m_discardCallback = cbDiscard;
    }

    void GrabResultMatcher::discardAll()
    {
        std::unique_lock<std::mutex> lock(m_mutex);
        discardAllImpl();
    }

    void GrabResultMatcher::discardAllImpl()
    {
        if (m_queueMatcher)
        {
            m_queueMatcher->discardAll();
        }
    }
    void GrabResultMatcher::onDiscard(size_t queueIndex, QueueItemBase* item)
    {
        if (m_discardCallback)
        {
            try
            {
                m_mutex.unlock();
                m_discardCallback(queueIndex, (dynamic_cast<QueueItemGrabResult&>(*item).grabResult()), m_pUserContext);
            }
            catch (...)
            {
            }
            m_mutex.lock();
        }
    }

    void GrabResultMatcher::onComplete(MatchingQueues&)
    {
        if (m_completionCallback)
        {
            try
            {
                m_mutex.unlock();
                m_completionCallback(*this, m_pUserContext);
            }
            catch (...)
            {
            }
            m_mutex.lock();
        }
    }



}

#ifndef PTPHELPER_NO_STREAM_OPERATORS
std::ostream& operator<<(std::ostream& out, const GenTLConsumerImplHelper::CameraInfo& info)
{
    out << "[" << info.strDisplayName << "]";
    return out;
}

std::ostream& operator<<(std::ostream& out, const blazePtpHelper::BlazeCamera& camera)
{
    out << camera.GetCameraInfo();
    return out;
}

std::ostream& operator<<(std::ostream& out, std::chrono::nanoseconds timestamp)
{
    const auto secs = std::chrono::duration_cast<std::chrono::seconds>(timestamp);
    const std::chrono::nanoseconds nsecs(timestamp - secs);
    out << secs.count() << "." << std::setw(9) << std::setfill('0') << nsecs.count();
    return out;
}
#endif