#pragma once
#include <stdint.h>
#include <memory>
#include <chrono>
#include <vector>
#include <mutex>
#include <deque>

#include <ConsumerImplHelper/BlazeCamera.h>

namespace blazePtpHelper
{
    using GrabResultArray = std::vector<GenTLConsumerImplHelper::GrabResult>;
    using BlazeCamera = GenTLConsumerImplHelper::CBlazeCamera;
    using BlazeCameraPtr = std::unique_ptr<BlazeCamera>;
    using BlazeCameraArray = std::vector<BlazeCameraPtr>;

    double calculateMaxFrameRate(const BlazeCameraArray&,
        bool interleaved,
        std::chrono::microseconds safetyMargin);

    bool openCamerasAndEnablePtp(BlazeCameraArray&);
    bool waitForMasterSlaveNegotation(const BlazeCameraArray&, std::chrono::seconds timeout);
    bool waitForClocksSynchronization(const BlazeCameraArray&,
        std::chrono::nanoseconds threshold,
        std::chrono::milliseconds window,
        std::chrono::seconds timeout);

    bool setupCamerasForSynchronousFreeRun(const BlazeCameraArray& cameras,
        std::chrono::microseconds safetyMargin = std::chrono::microseconds{ 400 });

    // ---------------------------------------------------------------------------
    void enablePtp(BlazeCamera& camera);

    void latchPtpStatus(BlazeCamera&);
    std::string getPtpStatus(BlazeCamera&);
    std::chrono::nanoseconds getOffsetFromMaster(BlazeCamera&);

    std::chrono::nanoseconds latchAndGetTimestamp(BlazeCamera&);

    std::chrono::microseconds getExposureTime(BlazeCamera&);
    std::chrono::microseconds getStartupTime(BlazeCamera&);
    std::chrono::microseconds getReaoutTime(BlazeCamera&);
    std::chrono::microseconds getResetTime(BlazeCamera&);

    //--------------------------------------------------------------------------
    extern std::ostream* debugStream;
    extern std::ostream* errorStream;

    inline std::ostream& debug() { return debugStream != nullptr ? *debugStream : std::clog; }
    inline std::ostream& error() { return errorStream != nullptr ? *errorStream : std::cerr; }
    void enableLogging(bool);

    std::string formatCameraName(const GenTLConsumerImplHelper::CameraInfo& info);
    std::string formatCameraName(const BlazeCamera& camera);

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
    class QueueItemBase;
    using QueueItemPtr = std::shared_ptr<QueueItemBase>;
    using ItemSet = std::vector<QueueItemPtr>;

    class MatchingQueues
    {
    public:
        using key_t = int64_t;
        using DiscardCallback = std::function<void(size_t queueIndex, QueueItemBase*)>;
        using CompletionCallback = std::function<void(MatchingQueues& queue)>;

        MatchingQueues(size_t nQueues, size_t maxQueueLen);
        MatchingQueues(const MatchingQueues&) = delete;
        MatchingQueues& operator=(const MatchingQueues&) = delete;
        ~MatchingQueues();

        void setDiscardCallback(DiscardCallback);
        void setCompletionCallback(CompletionCallback);

        void pushBack(size_t queueIndex, QueueItemBase* item);

        bool hasCompleteSet() const;
        ItemSet popSet();

        void discardAll();

    protected:
        void discardItem(size_t queueIndex, const QueueItemPtr& item);
        void discardItemsLessThan(key_t key);
        virtual void update();
        bool isThereNonEmptyQueue() const;
        bool isThereEmptyQueue() const;
        bool areAllQueuesEmpty() const { return !isThereNonEmptyQueue(); }
        std::pair<key_t, key_t> getMinMaxKey() const;

    private:
        enum class State
        {
            incomplete,
            complete
        } m_state{ State::incomplete };

    private:
        size_t m_nQueues = 2;
        size_t m_maxQueueLen = 3;

        DiscardCallback m_discardCallback;
        CompletionCallback m_completionCallback;

        using Queue = std::deque<QueueItemPtr>;
        using Queues = std::vector<Queue>;

        Queues m_queues;
    };

    class QueueItemBase
    {
    public:
        virtual ~QueueItemBase() = default;

        virtual MatchingQueues::key_t getKey() = 0;
    };

    using GrabResult = GenTLConsumerImplHelper::GrabResult;

    class GrabResultMatcher
    {
    public:
        using DiscardCallback = std::function<void(size_t cameraIndex, const GrabResult& grabResult, const void* pUserPtr)>;
        using CompletionCallback = std::function<void(GrabResultMatcher& matcher, const void* pUserPtr)>;

    public:
        GrabResultMatcher(size_t nCameras, size_t maxQueueLen);
        ~GrabResultMatcher();

        void addGrabResult(size_t cameraIdx, const GrabResult& GrabResultPtr);
        bool hasCompleteSet() const;
        GrabResultArray getGrabResults();

        void setCallbacks(CompletionCallback, DiscardCallback, const void* pUserPtr);

        void discardAll();

    protected:
        void discardAllImpl();

        void onDiscard(size_t queueIndex, QueueItemBase* item);
        void onComplete(MatchingQueues& queue);

    private:
        std::unique_ptr<MatchingQueues> m_queueMatcher;
        const void* m_pUserContext{ nullptr };
        CompletionCallback m_completionCallback;
        DiscardCallback    m_discardCallback;
        mutable std::mutex m_mutex;
    };
}

#ifndef PTPHELPER_NO_STREAM_OPERATORS
std::ostream& operator<<(std::ostream& out, const GenTLConsumerImplHelper::CameraInfo& info);
std::ostream& operator<<(std::ostream& out, const blazePtpHelper::BlazeCamera& camera);
std::ostream& operator<<(std::ostream& out, std::chrono::nanoseconds timestamp);
#endif
