/*
This sample illustrates how to operate several blaze cameras simultaneously
without their light sources interfering with each other.

Preventing the cameras from interfering with each other is done by enabling the
Synchronous Free Run feature and aligning the acquisition start times in a way 
that only one camera at a time sends out light.
If 2 cameras are found, the so-called interleaved free run mode will be 
configured. In this mode, the timing of two cameras is interleaved so that
the second camera sends out light when the first one performs the sensor 
readout. The interleaved mode doesn't reduce the possible frame rate.
If there are 3 or more cameras, the so-called consecutive free run mode 
will be used, where all cameras align there acquisitions so that each camera
acquires a complete frame while the other cameras are wating for their turn.

Synchronous Free Run requires that the camera clocks are synchronized using
PTP. That means that the cameras to be operated simultaneously have to be 
in the same subnet, i.e., a network switch is s required.

Refer to the blaze-101 product documentation for more details about PTP clock
synchronization and the Synchronous Free Run feature:
https://docs.baslerweb.com/synchronous-free-run-(blaze)
https://docs.baslerweb.com/precision-time-protocol-(blaze)

An alternative to operating multiple cameras using the Synchronous Free Run 
feature is using the Multi-Camera Channel feature, which doesn't require all 
cameras to be in the same subnet. Refer to the MultiCam/MultiChannelMultiThreaded
and the MultiCam/SingleThreaded samples to learn how to use the Multi-Camera 
Channel feature.

This sample uses the PtpHelper library located in the MultiCam/PtpHelper 
folder. In addition to a set of functions for performing the required setup 
steps for using synchronous free run, the library provides the GrabResultMatcher
class that assists in matching the captured frames from multiple cameras and 
retrieving matching frames as a set.

If you're planning to operate multiple cameras using Synchronous Fre Run and
PTP, the simplest approach is to include the MultiCam/PtpHelper/PtpHelper.h and 
MultiCam/PtpHelper/PtpHelper.cpp files in your project. 

OpenCV is only used for displaying the depth maps and is not required for using
the Synchronous Free Run and PTP features.

*/

#include <cstdlib>
#include <csignal>
#include <iostream>
#include <iomanip>
#include <chrono>
#include <algorithm>
#include <thread>
#include <mutex>
#include <atomic>
#include <condition_variable>

#if defined WITH_OPENCV
#  include <opencv2/opencv.hpp>
#endif

#include <ConsumerImplHelper/BlazeCamera.h>

#include <PtpHelper.h>

using namespace GenTLConsumerImplHelper;
namespace ptp = blazePtpHelper;
using namespace std::chrono_literals;

std::function<void()> stop_handler; // Functor to be called when CTRL+C is pressed.

// Signal handler that calls stop_handler when CTRL+C is pressed.
void signal_handler(int signal)
{
    if (signal == SIGINT && stop_handler) // CTRL+C
    {
        stop_handler();
    }
}

class Sample
{
public:
    bool run();
    void stop();

private:
    bool openCamerasAndEnablePtp();
    bool waitForClockSynchronization();
    bool setupCamerasForSynchronousFreeRun();
    bool startAcquisition();
    void grabLoop();  // Thread proceudre of the grab thead.
    void displayThreadFunction(); // Thread procedure of the display thread.

private:
    void onBuffersGrabbed(ptp::GrabResultMatcher& matcher, const void* pUserPtr);
    void onBufferDiscarded(size_t cameraIndex, const GrabResult& grabResult, const void* pUserPtr);

private:
    ptp::BlazeCameraArray   m_cameras;       // The blaze cameras.
    const size_t            m_nBuffers{ 5 }; // Number of buffers used for grabbing.
    std::unique_ptr<ptp::GrabResultMatcher> m_ptrGrabResultMatcher; // Used for matching frames by their frame IDs.
    std::atomic_bool        m_exitRequested{false};  // When true, a termination request is pending.
    std::thread             m_grabThread; // The thread grabbing frames from multiple cameras.
    std::thread             m_displayThread; // The thread that displays grabbed depth images.
    std::mutex              m_mtx;  // Used for synchronization with display thread.
    std::condition_variable m_cond; // used for synchronization with display thread.
};

/*
    Opens available cameras and enables the PTP feature.
*/
bool Sample::openCamerasAndEnablePtp()
{
    // Enumerate all blaze-101 cameras, create and open camera objects, 
    // enable PTP. The opened cameras are put into the m_cameras array.
    const bool result = ptp::openCamerasAndEnablePtp(m_cameras);

    if (m_cameras.empty())
    {
        std::cerr << "There is no camera to work with.";
        return false;
    }
    if (!result)
    {
        std::cerr << "Failed to open cameras or to enable PTP." << std::endl;
        return false;
    }
    if (1 == m_cameras.size())
    {
        std::cerr << "Warning: There is only one camera." << std::endl;
    }
    std::cout << "Using the following cameras:" << std::endl;
    for (const auto& camera : m_cameras)
    {
        std::cout << camera->GetCameraInfo().strDisplayName << std::endl;
    }
    return true;
}

/*
Waits until the clocks of all cameras in m_cameras are synchronized or a 
timeout expires.
*/
bool Sample::waitForClockSynchronization()
{
    // First, we have to wait until all the network devices have negotiated who is 
    // PTP master and who is PTP slave.
    const auto timeout = 30s;
    if (!ptp::waitForMasterSlaveNegotation(m_cameras, timeout))
    {
        std::cerr << "Waiting for master/slave negotiation timed out" << std::endl;
        return false;
    }
    
    // Then we have to wait until the clocks are synchronized, i.e., the estimated
    // offsets from the master clock don't exceed a given threshold for a given
    // period of time.
    const auto threshold = 1us;  // Maximum deviation of clocks.
    const auto window = 10s;     // The time period during which the deviations must be below the threshold.
    if (!ptp::waitForClocksSynchronization(m_cameras, threshold, window, timeout))
    {
        std::cerr << "Timeout expired before clocks were sufficiently synchronized." << std::endl;
        return false;
    }

    // Print the cameras' timestamps.
    // Note: Since we cannot simultaneously get the timestamps from all cameras
    // at the same time, there will be offsets between the printed timestamps, 
    // even if the clocks are perfectly synchronized.
    auto previousTimeStamp = std::chrono::nanoseconds(0);
    for (size_t i = 0; i < m_cameras.size(); ++i)
    {
        auto& camera = *m_cameras.at(i);
        const auto timeStamp = ptp::latchAndGetTimestamp(camera);
        std::cout << camera << " Timestamp = "
            << timeStamp;
        if (i > 0)
        {
            std::cout << " Delta = "
                << std::chrono::duration_cast<std::chrono::microseconds>(timeStamp - previousTimeStamp).count()
                << "us.";
        }
        std::cout << std::endl;
        previousTimeStamp = timeStamp;
    }

    return true;
}

/*
Set up all cameras to run synchronously. Each camera is started with a delay
to make sure that only one camera at a time is sending out light.
*/
bool Sample::setupCamerasForSynchronousFreeRun()
{
    // All subsequent steps assume that parameters like `ExposureTime or `FastMode`, 
    // which influence the frame timing,  won't be changed anymore after this function
    // has been called.
    // So, if you want to adjust these parameters, do it here.
    // Example: 
    for (auto& camera : m_cameras)
    {
        auto ptrExposureTime = camera->GetParameter<GenApi::CFloatPtr>("ExposureTime");
        ptrExposureTime->SetValue(ptrExposureTime->GetMax());
        camera->SetCameraParameterValue<bool>("FastMode", false);
    }

    //  For each camera, an unique acquisition start time will be set.
    //  Since the grab result matcher (see below) uses frame IDs for identifying
    //  frames that belong together, it is necessary to set start times to some
    //  point of time in the future. This ensures that the application has 
    //  enough time to fire up the grab thread *before* the start time is 
    //  reached and cameras start sending data.
    return ptp::setupCamerasForSynchronousFreeRun(m_cameras);
}

/*
Configure the cameras to send depth images and start acquisition.
*/
bool Sample::startAcquisition()
{
    // Prepare and start acquisition.
    for (auto &camera : m_cameras)
    {
        try
        {
            // In this sample, we grab depth data as range maps.
            camera->SetCameraParameterValue("ComponentSelector", "Range");
            camera->SetCameraParameterValue("PixelFormat", "Coord3D_C16");
            // Prepare for acquisition.
            camera->PrepareAcquisition(m_nBuffers);
            // Enqueue buffers to be grabbed. The memory is provided by the 
            // camera classes. Refer to the CustomGrabLoop sample to learn
            // how an application can provide the memory buffers.
            for (size_t i = 0; i < m_nBuffers; ++i)
            {
                camera->QueueBuffer(i);
            }
            // Start acquisition. 
            // First getting ready for grabbing buffers...
            camera->StartAcquisition();
            // ... then starting acquisition on the camera.
            camera->IssueAcquisitionStartCommand();
        }
        catch (const GenICam::GenericException &e)
        {
            std::cerr << *camera << " Failed to prepare acquisition: " << e.GetDescription() << std::endl;
            return false;
        }
    }
    return true;
}

/*
Callback for the grab result matcher.
As soon the grab result matcher has received matching frames from all cameras,  
it fires a callback. 
Note: This callback will be fired from the grab thread's thread context.
*/
void Sample::onBuffersGrabbed(ptp::GrabResultMatcher& /*matcher*/, const void* /*pUserPtr*/)
{
    m_cond.notify_one();  // Signal the display thread that there is work to do.
}

/* 
When there is too much network traffic causing a signifcant amount of network 
packets to be dropped, complete frames may be missed. If this happens, the
grab result matcher won't be able to provide a complete set of images and 
has to skip the buffers. For each buffer to be skipped, a callback will be 
called.
In this callback we will requeue the discarded buffers in the grabber's 
buffer queue to ensure that there are always enough buffers available to 
be grabbed.
*/
void Sample::onBufferDiscarded(size_t cameraIndex, const GrabResult& grabResult, const void* /*pUserPtr*/)
{
    auto& camera = *m_cameras.at(cameraIndex);
    try
    {
        std::cerr << camera
            << " Buffer skipped. Frame ID = " << grabResult.frameID << std::endl;
        camera.QueueBuffer(grabResult.hBuffer);
    }
    catch (GenICam::RuntimeException& e)
    {
        std::cerr << camera
            << " Failed to requeue buffer: " << e.GetDescription() << std::endl;
    }
}

/*
The grab loop run by the grab thread.
*/
void Sample::grabLoop()
{
    try
    {
        do
        {
            // Iterate over all cameras and get a grab result.
            for (size_t i = 0; i < m_cameras.size(); ++i)
            {
                auto& camera = m_cameras.at(i);
                GrabResult grabResult;
                auto grabTimeout = 100ms; 
                // Try to retrieve grab results from the camera until a grab
                // timeout occurs.
                do
                {
                    camera->GetGrabResult(grabResult, (uint32_t)grabTimeout.count());
                    if (m_exitRequested)
                    {
                        return; // Terminate grab loop.
                    }
                    if (grabResult.status == GrabResult::Timeout)
                    {
                        // We ignore timeouts and continue with the next camera.
                        break;
                    }
                    else if (grabResult.status != GrabResult::Ok)
                    {
                        std::cerr << *camera
                            << " Failed to grab image" << std::endl;
                        // Requeue buffer for grabbing.
                        camera->QueueBuffer(grabResult.hBuffer);
                    }
                    else
                    {
                        // Success. Pass the grab result to the grab result 
                        // matcher.
                        // The grab result matcher will fire a callback as soon
                        // as it has received matching frames from all cameras.
                        m_ptrGrabResultMatcher->addGrabResult(i, grabResult);
                    }
                    // Now try to retrieve grab results with zero timeout in 
                    // case grab results have queued up for some reason.
                    grabTimeout = 0ms;
                } while (true);
            }
        } while (!m_exitRequested);
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Exception in grab loop: " << e.GetDescription() << std::endl;
    }

    // Stop acquisition & clean up before terminating the thread.
    for (auto& camera : m_cameras)
    {
        try
        {
            camera->StopAcquisition();
            camera->FinishAcquisition();
        }
        catch (const GenICam::GenericException& e)
        {
            std::cerr << "Exception in grab loop when stopping acquisition: " 
                << e.GetDescription() << std::endl;
        }
    }
}

/* 
The thread consuming the grab results.
Here we just visualize depth maps (when OpenCV is available) and print out
timestamps of grabbed images.
*/
void Sample::displayThreadFunction()
{
#if defined WITH_OPENCV
    for (auto& camera : m_cameras)
    {
        // Prepare window for displaying the depth map.
        cv::namedWindow(camera->GetCameraInfo().strDisplayName, cv::WINDOW_AUTOSIZE);
    }
#endif

    do
    {
        // Check whether a matching set of results is available.
        if (!m_ptrGrabResultMatcher->hasCompleteSet())
        {
            // No result available. Wait until the grab thread lets us 
            // know that there is work to do.
            std::unique_lock<std::mutex> lock(m_mtx);
            m_cond.wait(lock, [this]() {
                return m_exitRequested || m_ptrGrabResultMatcher->hasCompleteSet();
            });
        }
        if (m_exitRequested)
        {
            // Meanwhile, termination was requested. 
            break;
        }

        // Process the matching images.
        // Note: When done, the buffers have to passed back to the camera
        // object.
        auto grabResults = m_ptrGrabResultMatcher->getGrabResults();

        auto previousTimeStamp = 0ns;
        for (size_t i = 0; i < grabResults.size(); ++i)
        {
            auto& camera = *m_cameras.at(i);
            const auto& grabResult = grabResults.at(i);

            // Access the data.
            BufferParts parts;
            camera.GetBufferParts(grabResult, parts);

#if defined WITH_OPENCV
            cv::Mat depth = cv::Mat((int)parts[0].height, (int)parts[0].width, CV_16UC1, parts[0].pData);
            cv::imshow(camera.GetCameraInfo().strDisplayName, depth);
#endif
            // grabResult.frameID is a frame counter.
            // Note: It's a 16-bit value, i.e., it "jumps back" from 65535 to
            // 1. It will never be zero.
            // grabResult.timeStamp indicates the camera's PTP timestamp at the moment
            // the camera started the acquisition of the frame.
            if (grabResult.frameID % 10 == 0)
            {
                std::cout << camera
                    << " FrameId = " << std::setw(4) << grabResult.frameID
                    << " Time stamp = " << std::chrono::nanoseconds(grabResult.timeStamp);
                auto currentTimeStamp = std::chrono::nanoseconds(grabResult.timeStamp);
                // Print the difference between the timestamp and the previous 
                // frame's timestamp.
                if (i > 0)
                {
                    std::cout << " Delta = "
                        << std::chrono::duration_cast<std::chrono::microseconds>(
                            currentTimeStamp - previousTimeStamp).count()
                        << "us";
                }
                std::cout << std::endl;
                previousTimeStamp = currentTimeStamp;
            }

            // When finished processing, requeue the buffer.
            // Note: By requeing a buffer, its ownership is transferred
            // back to the camera object. You must not access 
            // the data anymore since the buffers will be overwritten 
            // with new data.
            try
            {
                camera.QueueBuffer(grabResults.at(i).hBuffer);
            }
            catch (const GenICam::GenericException& e)
            {
                std::cerr << camera
                    << " Failed to requeue buffer: " << e.GetDescription() << std::endl;
            }
        }

#if defined WITH_OPENCV
        auto key = (char)cv::waitKey(1);  // Updates image windows and checks for key press.
        if (key == 'q' || key == 27)
        {
            m_exitRequested = true;
        }
#endif
    } while (!m_exitRequested);
}

bool Sample::run()
{
    m_exitRequested = false;

    try
    {
        /*
        Perform all required steps to set up available cameras for synchronous
        free run and to start acquisition.
        For each camera, a unique acquisition start time will be set.
        Since the grab result matcher (see below) uses frame IDs for identifying
        frames that belong together, it is necessary to set start times to some
        point of time in the future. This ensures that the application has 
        enough time to fire up the grab thread *before* the start time is
        reached and cameras start sending data.
        */
        if (!openCamerasAndEnablePtp()
        || !waitForClockSynchronization()
        || !setupCamerasForSynchronousFreeRun()
        || !startAcquisition() )
        {
            // One of the steps failed, exit.
            return false;
        }

#if defined WITH_OPENCV
        std::cout << "To exit, press CTRL+C, or press 'q' or 'ESC' when one of the image windows has the focus." << std::endl;
#else
        std::cout << "To exit, press CTRL+C." << std::endl;
#endif
        // Install signal handler for detecting CTRL+C.
        std::signal(SIGINT, signal_handler);
        stop_handler = [this]() { this->stop(); };

        // Initialize the grab result matcher.
        // For each camera, the grab result matcher maintains a queue where
        // it stores grab results. 
        // When there are matching grab results in all queues, a callback will
        // be fired.
        // The grab result matcher uses the frame IDs of the buffers to
        // match them.
        // 
        m_ptrGrabResultMatcher = std::make_unique<ptp::GrabResultMatcher>(
            m_cameras.size(), // number of cameras
            m_nBuffers - 1    // maximum n
        );  
        using namespace std::placeholders;
        m_ptrGrabResultMatcher->setCallbacks(
            std::bind(&Sample::onBuffersGrabbed, this, _1, _2),
            std::bind(&Sample::onBufferDiscarded, this, _1, _2, _3),
            nullptr);

        // Start the threads.
        m_displayThread = std::thread(&Sample::displayThreadFunction, this);
        m_grabThread = std::thread(&Sample::grabLoop, this);
        // Wait for the threads to terminate.
        // Termination will be requested by pressing CTRL+C
        // or, if OpenCV is available, pressing the q or ESC keys.
        m_grabThread.join();
        m_displayThread.join();
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Error occurred: " << e.GetDescription() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error occurred: " << e.what() << std::endl;
    }
    return true;
}

void Sample::stop()
{
    // Signal the threads to terminate.
    m_exitRequested = true;
    m_cond.notify_one();
}

int main()
{
    auto exitCode = EXIT_FAILURE;

    // If you don't want the PtpHelper library to write any
    // debug and progress information, you can disable the output:
    //ptp::enableLogging(false);
    try
    {
        CBlazeCamera::InitProducer();

        exitCode = Sample().run() ? EXIT_SUCCESS : EXIT_FAILURE;
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Error occurred: " << e.GetDescription() << std::endl;
    }

    // Clean up.
    if ( CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();

    return exitCode;
}
