/*
This sample illustrates how to save point clouds to file using
the PCL. The sample also shows how to load point clouds from file.
Using the PCL to save point clouds as binary files is more efficient than
writing text files "manually" as shown in the SavePointCloud sample.
*/

#include <chrono>

#include <pcl/point_types.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/cloud_viewer.h>

#include <ConsumerImplHelper/BlazeCamera.h>

using namespace GenTLConsumerImplHelper;
using namespace pcl;

// Typedefs for the PCL types used
typedef PointXYZRGB Point_t;
typedef PointCloud<Point_t> PointCloud_t;
typedef PointCloud_t::Ptr PointCloudPtr;

class Sample
{
public:
    int run();
    bool onImageGrabbed(GrabResult grabResult, BufferParts);

private:
    PointCloudPtr convertGrabResultToPointCloud(const BufferParts&);
    void loadAndShowPointClouds();

private:
    CBlazeCamera m_Camera;
    uint32_t     m_nPointCloudsSaved = 0;
    static const uint32_t c_countOfPointCloudsToSave = 10;
};

/*
Convert the Basler point cloud to a PCL point cloud.
*/
PointCloudPtr Sample::convertGrabResultToPointCloud(const BufferParts& imgParts)
{
    // An organized point cloud is used, i.e., for each camera pixel there is an entry 
    // in the data structure indicating the 3D coordinates calculated from that pixel.

    // Allocate PCL point cloud.
    const size_t width = imgParts[0].width;
    const size_t height = imgParts[0].height;
    PointCloudPtr ptrPointCloud(new PointCloud_t);
    ptrPointCloud->width = width;
    ptrPointCloud->height = height;
    ptrPointCloud->points.resize(width * height);
    ptrPointCloud->is_dense = false; // Indicates an organized point cloud.

    // Create a pointer to the 3D coordinates of the first point.
    // imgParts[0] always refers to the point cloud data.
    CBlazeCamera::Coord3D* pSrcPoint = (CBlazeCamera::Coord3D*) imgParts[0].pData;

    // Create a pointer to the intensity information, which is stored in the second buffer part.
    uint16_t* pIntensity = (uint16_t*)imgParts[1].pData;

    // Set the points.
    for (size_t i = 0; i < height * width; ++i, ++pSrcPoint, ++pIntensity)
    {
        // Set the X/Y/Z cordinates.
        Point_t& dstPoint = ptrPointCloud->points[i];

        dstPoint.x = pSrcPoint->x;
        dstPoint.y = pSrcPoint->y;
        dstPoint.z = pSrcPoint->z;

        // Use the intensity value of the pixel for coloring the point.
        dstPoint.r = dstPoint.g = dstPoint.b = (uint8_t)(*pIntensity >> 8);
    }

    return ptrPointCloud;
}

// Saves acquired point clouds.
bool Sample::onImageGrabbed(GrabResult grabResult, BufferParts parts)
{
    bool success = true;
    if (grabResult.status == GrabResult::Ok)
    {
        // Convert grab result to PCL point cloud.
        PointCloudPtr ptrPointCloud = convertGrabResultToPointCloud(parts);
        // Write point cloud to file.
        std::ostringstream filename;
        filename << m_nPointCloudsSaved << ".pcd";
        std::cout << "Saving " << filename.str() << std::endl;
        const auto t0 = std::chrono::high_resolution_clock::now();
        success = -1 != pcl::io::savePCDFileBinary(filename.str(), *ptrPointCloud);
        if (!success)
        {
            std::cerr << "Failed to save " << filename.str() << std::endl;
        }
        else
        {
            const auto duration = std::chrono::high_resolution_clock::now() - t0;
            std::cout << "Took " << std::chrono::duration_cast<std::chrono::microseconds>(duration).count() / 1000.0 
                << " ms to save point cloud." << std::endl;
        }
        m_nPointCloudsSaved++;
    }
    // Indicate to stop either if there was an error saving the point cloud or
    // when the specified number of point clouds has been saved.
    return success && m_nPointCloudsSaved < c_countOfPointCloudsToSave;
}

/*
Initialize the PCL CloudViewer. This function only gets called once.
For further information, see http://pointclouds.org/documentation/tutorials/cloud_viewer.html.
*/
void initCloudViewer(visualization::PCLVisualizer& viewer)
{
    // Set some properties of the visualizer.
    viewer.setBackgroundColor(0.1, 0.1, 0.1);
#if PCL_VERSION_COMPARE(>=,1,8,0)
    viewer.removeAllCoordinateSystems();
#endif
    // Set up the visualizer's camera.

    // The camera position.
    const double pos_x = 0; const double pos_y = 0; const double pos_z = -4000;

    // The point the camera is looking at.
    const double view_x = 0; const double view_y = 0; const double view_z = 3000;

    // Defines the view-up direction of the camera ( up_y == -1 --> y axis of the blaze camera points down )
    const double up_x = 0; const double up_y = -1; const double up_z = 0;
    viewer.setCameraPosition(pos_x, pos_y, pos_z, view_x, view_y, view_z, up_x, up_y, up_z);

    // Set the size of the points rendered.
    viewer.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3);
}

// Loads point clouds from file and visualizes them until the user stops the
// visualization.
void Sample::loadAndShowPointClouds()
{
    bool viewerInitialized = false;
    int result = 0;
    auto viewer = std::make_shared<pcl::visualization::CloudViewer>("CloudViewer");
    int idx = 0;

    do
    {   
        std::ostringstream filename;
        filename << idx << ".pcd";
        std::cout << "Loading " << filename.str() << std::endl;
        auto ptrPointCloud = PointCloudPtr(new PointCloud_t);
        const auto t0 = std::chrono::high_resolution_clock::now();
        auto result = pcl::io::loadPCDFile(filename.str(), *ptrPointCloud);
        const auto duration = std::chrono::high_resolution_clock::now() - t0;
        std::cout << "Took "
            << std::chrono::duration_cast<std::chrono::microseconds>(duration).count() / 1000.0
            << " ms to load point cloud." << std::endl;
        if (-1 != result)
        {
            viewer->showCloud(ptrPointCloud);
            if (!viewerInitialized)
            {
                viewer->runOnVisualizationThreadOnce(initCloudViewer);
                viewerInitialized = true;
            }
        }
        else
        {
            std::cerr << "Failed to load point cloud '" << filename.str() << "'" << std::endl;
        }
        if (++idx >= m_nPointCloudsSaved)
            idx = 0;
    } while (result != -1 && !viewer->wasStopped());
}

int Sample::run()
{
    try
    {
        // Open the first camera found, i.e., establish a connection to the camera device. 
        m_Camera.OpenFirstCamera();

        std::cout << "Connected to camera " << m_Camera.GetCameraInfo().strDisplayName << std::endl;

        // Enable 3D (point cloud) data & intensity data.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Range");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Coord3D_ABC32f");

        m_Camera.SetCameraParameterValue("ComponentSelector", "Intensity");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Mono16");

        // Acquire data and save point clouds until the onImageGrabbed callback indicates to stop acquisition. 
        // 3 buffers are used (round-robin).
        m_Camera.GrabContinuous(3, 1000, this, &Sample::onImageGrabbed);

        if (m_nPointCloudsSaved > 0)
        {
            std::cout << "Press 'q' to exit." << std::endl;
            loadAndShowPointClouds();
        }

        // Clean-up
        m_Camera.Close();
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used 
        // to check if the device is still connected.
        if (m_Camera.IsOpen() && !m_Camera.IsConnected())
        {
            std::cerr << "Camera has been removed." << std::endl;
        }

        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try
    {
        CBlazeCamera::InitProducer();

        Sample processing;

        exitCode = processing.run();
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CToFCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}
