/*
This sample illustrates how to grab images from a blaze camera and how to access
the image, depth data, and confidence map.
The grab loop provided by the CBlazeCamera class is used. Different from the 
FirstSample sample program, a class member
function is passed as grab callback to CBlazeCamera::GrabContinuous().

Further Reading:

The 'FirstSample' sample illustrates how a lambda function instead of a member 
function can be used as a grab callback.
The 'CameraConfiguration' sample illustrates in more detail how to configure a 
blaze camera.
The 'CustomGrabLoop' sample illustrates more advanced use cases like how to use
user-provided buffers and how an application can provide its own grab loop.
*/

#include <iostream>
#include <iomanip> 
#include <ConsumerImplHelper/BlazeCamera.h>

using namespace GenTLConsumerImplHelper;

class Sample
{
public:
    int run();

    bool onImageGrabbed(GrabResult grabResult, BufferParts);

private:
    CBlazeCamera m_Camera;
    int          m_nBuffersGrabbed;
};

int Sample::run()
{
    m_nBuffersGrabbed = 0;

    try
    {
        // Open the first camera found, i.e., establish a connection to the camera device. 
        m_Camera.OpenFirstCamera();

        /*  CBlazeCamera::OpenFirstCamera() is a shortcut for the following sequence:
            CameraList cameras = CBlazeCamera::EnumerateCameras();
            CameraInfo camInfo = *cameras.begin();
            m_Camera.Open(camInfo);

            If there are multiple cameras connected and you want to open a specific one, use
            the CBlazeCamera::Open( CameraInfoKey, string ) method.

            Example: Open a camera using its IP address
            CBlazeCamera::Open( IpAddress, "192.168.0.2" );

            Instead of the IP address, any other property of the CameraInfo struct can be used,
            e.g., the serial number:

            CBlazeCamera::Open( SerialNumber, "23167572" );
        */

        std::cout << "Connected to camera " << m_Camera.GetCameraInfo().strDisplayName << std::endl;

        // Enable 3D (point cloud) data, intensity data, and confidence data. 

        // Enable range data.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Range");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        // Range information can be sent either as a 16-bit gray value image or as 3D coordinates (point cloud). 
        // For this sample, we want to acquire 3D coordinates.
        // Note: To change the format of an image component, the Component Selector must first be set to the component
        // you want to configure (see above).
        // To use 16-bit integer depth information, choose "Coord3D_C16" instead of "Coord3D_ABC32f". For more details 
        // about how to process 16-bit depth images, refer to the GrabDepthMap sample.
        m_Camera.SetCameraParameterValue("PixelFormat", "Coord3D_ABC32f");

       
        // Enable intensity image.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Intensity");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Mono16");

        // Enable confidence map.
        m_Camera.SetCameraParameterValue("ComponentSelector", "Confidence");
        m_Camera.SetCameraParameterValue("ComponentEnable", true);
        m_Camera.SetCameraParameterValue("PixelFormat", "Confidence16"); // Request 16bit per Pixel

        // Configure the value used for identifying missing depth information.
        m_Camera.SetCameraParameterValue("Scan3dCoordinateSelector", "CoordinateC"); // Choose z-axis
        // z-coordinate will be set to this value if there is no valid depth data available
        m_Camera.SetCameraParameterValue("Scan3dInvalidDataValue", 0.0);

        // Acquire images until the onImageGrabbed callback indicates to stop acquisition. 
        // 5 buffers are used (round-robin).
        m_Camera.GrabContinuous(5, 1000, this, &Sample::onImageGrabbed);

        // Clean-up
        m_Camera.Close();
    }
    catch (const GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << e.GetDescription() << std::endl;
        // After successfully opening the camera, the IsConnected method can be used 
        // to check if the device is still connected.
        if (m_Camera.IsOpen() && !m_Camera.IsConnected())
        {
            std::cerr << "Camera has been removed." << std::endl;
        }

        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}

bool Sample::onImageGrabbed(GrabResult grabResult, BufferParts parts)
{
    if (grabResult.status == GrabResult::Timeout)
    {
        std::cerr << "Timeout occurred. Acquisition stopped." << std::endl;
        // The timeout might be caused by a removal of the camera. Check if the camera
        // is still connected.
        if (!m_Camera.IsConnected())
        {
            std::cerr << "Camera has been removed." << std::endl;
        }

        return false; // Indicate to stop acquisition
    }
    m_nBuffersGrabbed++;
    if (grabResult.status != GrabResult::Ok)
    {
        std::cerr << "Image " << m_nBuffersGrabbed << " was not grabbed." << std::endl;
    }
    else
    {
        // Retrieve the values for the center pixel
        const int width = (int)parts[0].width;
        const int height = (int)parts[0].height;
        const int x = (int)(0.5 * width);
        const int y = (int)(0.5 * height);
        CBlazeCamera::Coord3D* p3DCoordinate = (CBlazeCamera::Coord3D*) parts[0].pData + y * width + x;
        uint16_t* pIntensity = (uint16_t*)parts[1].pData + y * width + x;
        uint16_t* pConfidence = (uint16_t*)parts[2].pData + y * width + x;

        std::cout << "Center pixel of image " << std::setw(2) << m_nBuffersGrabbed << ": ";
        std::cout.setf(std::ios_base::fixed);
        std::cout.precision(1);
        if (p3DCoordinate->z != 0)
            std::cout << "x=" << std::setw(6) << p3DCoordinate->x << " y=" << std::setw(6) << p3DCoordinate->y << " z=" << std::setw(6) << p3DCoordinate->z;
        else
            std::cout << "x=   n/a y=   n/a z=   n/a";
        std::cout << " intensity=" << std::setw(5) << *pIntensity << " confidence=" << std::setw(5) << *pConfidence << "\n";
    }
    return m_nBuffersGrabbed < 10; // Indicate to stop acquisition when 10 buffers are grabbed
}

int main(int argc, char* argv[])
{
    int exitCode = EXIT_SUCCESS;

    try
    {
        CBlazeCamera::InitProducer();

        exitCode = Sample().run();
    }
    catch (GenICam::GenericException& e)
    {
        std::cerr << "Exception occurred: " << std::endl << e.GetDescription() << std::endl;
        exitCode = EXIT_FAILURE;
    }

    // Release the GenTL producer and all of its resources. 
    // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
    // class has been called. The destructor may require resources which may not
    // be available anymore after TerminateProducer() has been called.
    if (CBlazeCamera::IsProducerInitialized())
        CBlazeCamera::TerminateProducer();  // Won't throw any exceptions

    std::cout << std::endl << "Press Enter to exit." << std::endl;
    while (std::cin.get() != '\n');

    return exitCode;
}
